# -*- coding: utf-8 -*-
"""Test suite for axonapi.api.assets."""
import pytest


def load_test_data(apiobj):
    """Pass."""
    apiobj.TEST_DATA = getattr(apiobj, "TEST_DATA", {})

    if not apiobj.TEST_DATA.get("fields_map"):
        apiobj.TEST_DATA["fields_map"] = fields_map = apiobj.fields.get()

    if not apiobj.TEST_DATA.get("assets"):
        apiobj.TEST_DATA["assets"] = apiobj.get(max_rows=4000, fields_map=fields_map)

    return apiobj


class LabelsPrivate:
    """Pass."""

    def test_private_get(self, apiobj):
        """Pass."""
        fields = apiobj.labels._get()
        assert isinstance(fields, list)
        for x in fields:
            assert isinstance(x, str)

    def test_private_add_get_remove(self, apiobj):
        """Pass."""
        labels = ["badwolf1", "badwolf2"]

        # get a single asset to add a label to
        asset = apiobj.TEST_DATA["assets"][0]
        asset_id = asset["internal_axon_id"]

        # add the label to the asset
        add_label_result = apiobj.labels._add(labels=labels, ids=[asset_id])
        assert add_label_result == 1

        # re-get the asset and check that it has the label
        assets_added = apiobj.get_by_values(
            values=labels,
            field="labels",
            fields="labels",
            fields_map=apiobj.TEST_DATA["fields_map"],
        )
        assets_added_ids = [x["internal_axon_id"] for x in assets_added]
        assert asset_id in assets_added_ids

        # check the each label has been added
        for x in assets_added:
            for label in labels:
                assert label in x["labels"]

        # check that the label is in all the labels on the system
        all_labels_post_add = apiobj.labels._get()
        assert isinstance(all_labels_post_add, list)

        for label in all_labels_post_add:
            assert isinstance(label, str)

        for label in labels:
            assert label in all_labels_post_add

        # remove the label from an asset
        remove_label_result = apiobj.labels._remove(labels=labels, ids=assets_added_ids)
        assert remove_label_result >= 1

        # re-get the asset and check that it has the label
        assets_removed = apiobj.get_by_values(
            values=labels,
            field="labels",
            fields="labels",
            fields_map=apiobj.TEST_DATA["fields_map"],
        )
        assert not assets_removed

        # check that the label is not in all the labels on the system
        all_labels_post_remove = apiobj.labels._get()
        assert isinstance(all_labels_post_remove, list)

        for label in labels:
            assert label not in all_labels_post_remove


class LabelsPublic:
    """Pass."""

    def test_get(self, apiobj):
        """Pass."""
        fields = apiobj.labels.get()
        assert isinstance(fields, list)
        for x in fields:
            assert isinstance(x, str)

    def test_add_get_remove(self, apiobj):
        """Pass."""
        labels = ["badwolf1", "badwolf2"]

        # get a single asset to add a label to
        asset = apiobj.TEST_DATA["assets"][0]
        asset_id = asset["internal_axon_id"]

        # add the label to the asset
        add_label_result = apiobj.labels.add(labels=labels, rows=[asset])
        assert add_label_result == 1

        # re-get the asset and check that it has the label
        assets_added = apiobj.get_by_values(
            values=labels,
            field="labels",
            fields="labels",
            fields_map=apiobj.TEST_DATA["fields_map"],
        )
        assets_added_ids = [x["internal_axon_id"] for x in assets_added]
        assert asset_id in assets_added_ids

        # check the each label has been added
        for x in assets_added:
            for label in labels:
                assert label in x["labels"]

        # check that the label is in all the labels on the system
        all_labels_post_add = apiobj.labels.get()
        assert isinstance(all_labels_post_add, list)

        for label in all_labels_post_add:
            assert isinstance(label, str)

        for label in labels:
            assert label in all_labels_post_add

        # remove the label from an asset
        remove_label_result = apiobj.labels.remove(labels=labels, rows=assets_added)
        assert remove_label_result >= 1

        # re-get the asset and check that it has the label
        assets_removed = apiobj.get_by_values(
            values=labels,
            field="labels",
            fields="labels",
            fields_map=apiobj.TEST_DATA["fields_map"],
        )
        assert not assets_removed

        # check that the label is not in all the labels on the system
        all_labels_post_remove = apiobj.labels.get()
        assert isinstance(all_labels_post_remove, list)

        for label in labels:
            assert label not in all_labels_post_remove


class TestLabelsDevices(LabelsPrivate, LabelsPublic):
    """Pass."""

    @pytest.fixture(scope="class")
    def apiobj(self, api_devices):
        """Pass."""
        return load_test_data(api_devices)


class TestLabelsUsers(LabelsPrivate, LabelsPublic):
    """Pass."""

    @pytest.fixture(scope="class")
    def apiobj(self, api_users):
        """Pass."""
        return load_test_data(api_users)



# -*- coding: utf-8 -*-
"""Command line interface for Axonius API Client."""
from ...context import CONTEXT_SETTINGS, click
from ...options import AUTH, NODE_CNX, add_options
from .grp_common import ID_CNX

OPTIONS = [
    *AUTH,
    *NODE_CNX,
    ID_CNX,
]


@click.command(name="delete-by-id", context_settings=CONTEXT_SETTINGS)
@add_options(OPTIONS)
@click.pass_context
def cmd(ctx, url, key, secret, **kwargs):
    """Delete a connection by ID."""
    client = ctx.obj.start_client(url=url, key=key, secret=secret)

    with ctx.obj.exc_wrap(wraperror=ctx.obj.wraperror):
        client.adapters.cnx.delete_by_id(**kwargs)

    ctx.obj.echo_ok(f"Connection deleted!")



# -*- coding: utf-8 -*-
"""API models package."""
from . import actions, enforcements
from .actions import RunAction
from .enforcements import Enforcements

__all__ = (
    "Enforcements",
    "RunAction",
    "enforcements",
    "actions",
)



# -*- coding: utf-8 -*-
"""Command line interface for Axonius API Client."""
from ...tools import json_dump
from ..context import CONTEXT_SETTINGS, click
from ..options import AUTH, INPUT_FILE, NODE, add_options
from .grp_common import CONFIG_EXPORT, CONFIG_TYPE, config_export

OPTIONS = [*AUTH, CONFIG_EXPORT, CONFIG_TYPE, *NODE, INPUT_FILE]


@click.command(name="config-update-from-json", context_settings=CONTEXT_SETTINGS)
@add_options(OPTIONS)
@click.pass_context
def cmd(ctx, url, key, secret, export_format, input_file, **kwargs):
    """Set adapter advanced settings from a JSON file."""
    client = ctx.obj.start_client(url=url, key=key, secret=secret)
    contents = ctx.obj.read_stream_json(stream=input_file, expect=dict)
    kwargs["kwargs_config"] = contents

    with ctx.obj.exc_wrap(wraperror=ctx.obj.wraperror):
        rows = client.adapters.config_update(**kwargs)
    ctx.obj.echo_ok(f"Updated adapter with config:\n{json_dump(contents)}")
    config_export(ctx=ctx, rows=rows, export_format=export_format)



# -*- coding: utf-8 -*-
"""HTTP client."""
import logging
import warnings
from urllib.parse import urlparse, urlunparse

import requests

from .constants import (LOG_LEVEL_HTTP, MAX_BODY_LEN, REQUEST_ATTR_MAP,
                        RESPONSE_ATTR_MAP, TIMEOUT_CONNECT, TIMEOUT_RESPONSE)
from .exceptions import HttpError
from .logs import get_obj_log, set_log_level
from .tools import join_url, json_reload, listify, path_read
from .version import __version__

InsecureRequestWarning = requests.urllib3.exceptions.InsecureRequestWarning


class Http:
    """HTTP client wrapper around :obj:`requests.Session`."""

    def __init__(
        self,
        url,
        connect_timeout=TIMEOUT_CONNECT,
        response_timeout=TIMEOUT_RESPONSE,
        certpath=None,
        certwarn=True,
        certverify=False,
        cert_client_both=None,
        cert_client_cert=None,
        cert_client_key=None,
        http_proxy=None,
        https_proxy=None,
        save_last=True,
        save_history=False,
        log_level=LOG_LEVEL_HTTP,
        log_level_urllib="warning",
        log_request_attrs=None,
        log_response_attrs=None,
        log_request_body=False,
        log_response_body=False,
    ):
        """HTTP client wrapper around :obj:`requests.Session`.

        Notes:
            * If certpath is supplied, certverify is ignored
            * private key supplied to cert_client_key or cert_client_both
              can **NOT** be password encrypted

        Args:
            url (:obj:`str` or :obj:`ParserUrl`): URL to connect to
            connect_timeout (:obj:`int`, optional):
                default :data:`TIMEOUT_CONNECT` - seconds to
                wait for connections to open to :attr:`url`
            response_timeout (:obj:`int`, optional):
                default :data:`TIMEOUT_RESPONSE` - seconds to
                wait for responses from :attr:`url`
            certpath (:obj:`str` or :obj:`pathlib.Path`, optional): default ``None`` -
                path to CA bundle file to use when verifing certs offered by :attr:`url`
                instead of the system CA bundle
            certwarn (:obj:`bool`, optional): default ``True`` - show warnings from
                requests about certs offered by :attr:`url` that are self signed:

                * if ``True`` show warning only the first time it happens
                * if ``False`` never show warning
                * if ``None`` show warning every time it happens
            certverify (:obj:`bool`, optional): default ``False`` - control
                validation of certs offered by :attr:`url`:

                * if ``True`` raise exception if cert is invalid/self-signed
                * if ``False`` only raise exception if cert is invalid
            cert_client_both (:obj:`str` or :obj:`pathlib.Path`, optional):
                default ``None`` - path to cert file containing both the private key and
                cert to offer to :attr:`url`
            cert_client_cert (:obj:`str` or :obj:`pathlib.Path`, optional):
                default ``None`` - path to cert file to offer to :attr:`url`
                (*must also supply cert_client_key*)
            cert_client_key (:obj:`str` or :obj:`pathlib.Path`, optional):
                default ``None`` - path to private key file for cert_client_cert to
                offer to :attr:`url` (*must also supply cert_client_cert*)
            http_proxy (:obj:`str`, optional): default ``None`` - proxy to use
                when making http requests to :attr:`url`
            https_proxy (:obj:`str`, optional): default ``None`` - proxy to use
                when making https requests to :attr:`url`
            save_last (:obj:`bool`, optional): default ``True`` -

                * if ``True`` save request to :attr:`LAST_REQUEST` and
                  response to :attr:`LAST_RESPONSE`
                * if ``False`` do not save request to :attr:`LAST_REQUEST` and
                  response to :attr:`LAST_RESPONSE`
            save_history (:obj:`bool`, optional): default ``True`` -

                * if ``True`` append responses to :attr:`HISTORY`
                * if ``False`` do not append responses to :attr:`HISTORY`
            log_level (:obj:`str`):
              default :data:`axonius_api_client.LOG_LEVEL_HTTP` -
              logging level to use for this objects logger
            log_level_urllib (:obj:`str`): default ``"warning"`` -
              logging level to use for this urllib logger
            log_request_attrs (:obj:`bool`): default ``None`` - control logging
              of request attributes:

              * if ``True``, log request attributes defined in
                :data:`axonius_api_client.LOG_REQUEST_ATTRS_VERBOSE`
              * if ``False``, log request attributes defined in
                :data:`axonius_api_client.LOG_REQUEST_ATTRS_BRIEF`
              * if ``None``, do not log any request attributes
            log_response_attrs (:obj:`bool`): default ``None`` - control logging
              of response attributes:

              * if ``True``, log response attributes defined in
                :data:`axonius_api_client.LOG_RESPONSE_ATTRS_VERBOSE`
              * if ``False``, log response attributes defined in
                :data:`axonius_api_client.LOG_RESPONSE_ATTRS_BRIEF`
              * if ``None``, do not log any response attributes
            log_request_body (:obj:`bool`): default ``False`` - control logging
              of request bodies:

              * if ``True``, log request bodies
              * if ``False``, do not log request bodies
            log_response_body (:obj:`bool`): default ``False`` - control logging
              of response bodies:

              * if ``True``, log response bodies
              * if ``False``, do not log response bodies

        Raises:
            :exc:`HttpError`: if either cert_client_cert or cert_client_key
                are supplied, and the other is not supplied

            :exc:`HttpError`: if any of cert_path, cert_client_cert,
                cert_client_key, or cert_client_both are supplied and the file does
                not exist
        """
        self.LOG = get_obj_log(obj=self, level=log_level)
        """:obj:`logging.Logger`: Logger for this object."""

        if isinstance(url, ParserUrl):
            self.URLPARSED = url
        else:
            self.URLPARSED = ParserUrl(url=url, default_scheme="https")

        self.url = self.URLPARSED.url
        """:obj:`str`: URL to connect to"""

        self.LAST_REQUEST = None
        """:obj:`requests.PreparedRequest`: last request sent"""

        self.LAST_RESPONSE = None
        """:obj:`requests.Response`: last response received"""

        self.HISTORY = []
        """:obj:`list` of :obj:`requests.Response`: all responses received."""

        self.SAVE_LAST = save_last
        """:obj:`bool`: save requests to :attr:`LAST_REQUEST` and responses
        to :attr:`LAST_RESPONSE`"""

        self.SAVEHISTORY = save_history
        """:obj:`bool`: Append all responses to :attr:`HISTORY`"""

        self.CONNECT_TIMEOUT = connect_timeout
        """:obj:`int`: seconds to wait for connections to open to :attr:`url`"""

        self.RESPONSE_TIMEOUT = response_timeout
        """:obj:`int`: seconds to wait for responses from :attr:`url`"""

        self.session = requests.Session()
        """:obj:`requests.Session`: session object to use"""

        self.LOG_REQUEST_BODY = log_request_body
        """:obj:`bool`: Log the full request body."""

        self.LOG_RESPONSE_BODY = log_response_body
        """:obj:`bool`: Log the full response body."""

        self.log_request_attrs = log_request_attrs
        self.log_response_attrs = log_response_attrs

        self.session.proxies = {}
        self.session.proxies["https"] = https_proxy
        self.session.proxies["http"] = http_proxy

        if certpath:
            path_read(obj=certpath, binary=True)
            self.session.verify = certpath
        else:
            self.session.verify = certverify

        if cert_client_both:
            path_read(obj=cert_client_both, binary=True)
            self.session.cert = str(cert_client_both)
        elif cert_client_cert or cert_client_key:
            if not all([cert_client_cert, cert_client_key]):
                error = (
                    "You must supply both a 'cert_client_cert' and 'cert_client_key'"
                    " or use 'cert_client_both'!"
                )
                raise HttpError(error)

            path_read(obj=cert_client_cert, binary=True)
            path_read(obj=cert_client_key, binary=True)
            self.session.cert = (str(cert_client_cert), str(cert_client_key))

        if certwarn is True:
            warnings.simplefilter("once", InsecureRequestWarning)
        elif certwarn is False:
            warnings.simplefilter("ignore", InsecureRequestWarning)

        urllog = logging.getLogger("urllib3.connectionpool")
        set_log_level(obj=urllog, level=log_level_urllib)

    def __call__(
        self,
        path=None,
        route=None,
        method="get",
        data=None,
        params=None,
        headers=None,
        json=None,
        files=None,
        # fmt: off
        **kwargs
        # fmt: on
    ):
        """Create, prepare, and then send a request using :attr:`session`.

        Args:
            path (:obj:`str`, optional): default ``None`` - path to append to
                :attr:`url`
            route (:obj:`str`, optional): default ``None`` - route to append to
                :attr:`url`
            method (:obj:`str`, optional): default ``"get"`` - method to use
            data (:obj:`str`, optional): default ``None`` - body to send
            params (:obj:`dict`, optional): default ``None`` - parameters to url encode
            headers (:obj:`dict`, optional): default ``None`` - headers to send
            json (:obj:`dict`, optional): default ``None`` - obj to encode as json
            files (:obj:`tuple` of :obj:`tuple`, optional): default ``None`` - files to
                send
            **kwargs:
                overrides for object attributes

                * connect_timeout (:obj:`int`): default :attr:`CONNECT_TIMEOUT` -
                  seconds to wait for connection to open to :attr:`url`
                * response_timeout (:obj:`int`): default :attr:`RESPONSE_TIMEOUT` -
                  seconds to wait for for response from :attr:`url`
                * proxies (:obj:`dict`): default ``None`` -
                  use custom proxies instead of proxies defined in :attr:`session`
                * verify (:obj:`bool` or :obj:`str`): default ``None`` - use custom
                  verification of cert offered by :attr:`url` instead of verification
                  defined in :attr:`session`
                * cert (:obj:`str`): default ``None`` - use custom
                  client cert to offer to :attr:`url` cert defined in :attr:`session`

        Returns:
            :obj:`requests.Response`: raw response object

        """
        url = join_url(self.url, path, route)

        headers = headers or {}
        headers.setdefault("User-Agent", self.user_agent)

        request = requests.Request(
            url=url,
            method=method,
            data=data,
            headers=headers,
            params=params,
            json=json,
            files=files or [],
        )
        prepped_request = self.session.prepare_request(request=request)
        prepped_request.body_size = len(prepped_request.body or "")

        if self.SAVE_LAST:
            self.LAST_REQUEST = prepped_request

        if self.log_request_attrs:
            lattrs = ", ".join(self.log_request_attrs).format(request=prepped_request)
            self.LOG.debug(f"REQUEST ATTRS: {lattrs}")

        send_args = self.session.merge_environment_settings(
            url=prepped_request.url,
            proxies=kwargs.get("proxies", {}),
            stream=kwargs.get("stream", None),
            verify=kwargs.get("verify", None),
            cert=kwargs.get("cert", None),
        )

        send_args["request"] = prepped_request
        send_args["timeout"] = (
            kwargs.get("connect_timeout", self.CONNECT_TIMEOUT),
            kwargs.get("response_timeout", self.RESPONSE_TIMEOUT),
        )

        if self.LOG_REQUEST_BODY:
            self.log_body(body=prepped_request.body, body_type="REQUEST")

        response = self.session.send(**send_args)
        response.body_size = len(response.text or "")

        if self.SAVE_LAST:
            self.LAST_RESPONSE = response

        if self.SAVEHISTORY:
            self.HISTORY.append(response)

        if self.log_response_attrs:
            lattrs = ", ".join(self.log_response_attrs).format(response=response)
            self.LOG.debug(f"RESPONSE ATTRS: {lattrs}")

        if self.LOG_RESPONSE_BODY:
            self.log_body(body=response.text, body_type="RESPONSE")

        return response

    def __str__(self):
        """Show object info.

        Returns:
            :obj:`str`
        """
        return "{c.__module__}.{c.__name__}(url={url!r})".format(
            c=self.__class__, url=self.url
        )

    def __repr__(self):
        """Show object info.

        Returns:
            :obj:`str`
        """
        return self.__str__()

    @property
    def user_agent(self):
        """Value to use in User-Agent header.

        Returns:
            :obj:`str`: user agent string
        """
        return f"{__name__}.{self.__class__.__name__}/{__version__}"

    @property
    def log_request_attrs(self):
        """Get the request attributes that should be logged."""
        return self._get_log_attrs("request")

    @log_request_attrs.setter
    def log_request_attrs(self, value):
        """Set the request attributes that should be logged."""
        attr_map = REQUEST_ATTR_MAP
        attr_type = "request"
        self._set_log_attrs(attr_map=attr_map, attr_type=attr_type, value=value)

    @property
    def log_response_attrs(self):
        """Get the response attributes that should be logged."""
        return self._get_log_attrs("response")

    @log_response_attrs.setter
    def log_response_attrs(self, value):
        """Set the response attributes that should be logged."""
        attr_map = RESPONSE_ATTR_MAP
        attr_type = "response"
        self._set_log_attrs(attr_map=attr_map, attr_type=attr_type, value=value)

    def _get_log_attrs(self, attr_type):
        return getattr(self, "_LOG_ATTRS", {}).get(attr_type, [])

    def _set_log_attrs(self, attr_map, attr_type, value):
        if not hasattr(self, "_LOG_ATTRS"):
            self._LOG_ATTRS = {"response": [], "request": []}

        value = [x.lower().strip() for x in listify(value)]

        if not value:
            self._LOG_ATTRS[attr_type] = []
            return

        log_attrs = self._LOG_ATTRS[attr_type]

        if "all" in value:
            for k, v in attr_map.items():
                entry = f"{k}={v}"
                if entry not in log_attrs:
                    log_attrs.append(entry)
            return

        for item in value:
            if item in attr_map:
                value = attr_map[item]
                entry = f"{item}={value}"
                if entry not in log_attrs:
                    log_attrs.append(entry)

    def log_body(self, body, body_type):
        """Pass."""
        body = body or ""
        body = json_reload(obj=body, error=False, trim=MAX_BODY_LEN)
        self.LOG.debug(f"{body_type} BODY:\n{body}")


class ParserUrl:
    """Parse a URL and ensure it has the neccessary bits."""

    def __init__(self, url, default_scheme="https"):
        """Parse a URL and ensure it has the neccessary bits.

        Args:
            url (:obj:`str`): URL to parse
            default_scheme (:obj:`str`, optional): default ``"https"`` - default
                scheme to use if url does not contain a scheme

        Raises:
            :exc:`HttpError`:
                if parsed URL winds up without a hostname, port, or scheme.

        """
        self._init_url = url
        """:obj:`str`: initial URL provided"""

        self._init_scheme = default_scheme
        """:obj:`str`: default scheme provided"""

        self._init_parsed = urlparse(url)
        """:obj:`urllib.parse.ParseResult`: first pass of parsing URL"""

        self.parsed = self.reparse(
            parsed=self._init_parsed, default_scheme=default_scheme
        )
        """:obj:`urllib.parse.ParseResult`: second pass of parsing URL"""

        for part in ["hostname", "port", "scheme"]:
            if not getattr(self.parsed, part, None):
                error = (
                    f"Parsed URL into {self.parsed_str!r} and no {part!r} provided"
                    f" in URL {url!r}"
                )
                raise HttpError(error)

    def __str__(self):
        """Show object info.

        Returns:
            :obj:`str`
        """
        cls = self.__class__
        return f"{cls.__module__}.{cls.__name__}({self.parsed_str})"

    def __repr__(self):
        """Show object info.

        Returns:
            :obj:`str`
        """
        return self.__str__()

    @property
    def hostname(self):
        """Hostname part from :attr:`ParserUrl.parsed`.

        Returns:
            :obj:`str`: hostname value
        """
        return self.parsed.hostname

    @property
    def port(self):
        """Port part from :attr:`ParserUrl.parsed`.

        Returns
            :obj:`int`: port value
        """
        return int(self.parsed.port)

    @property
    def scheme(self):
        """Scheme part from :attr:`ParserUrl.parsed`.

        Returns:
            :obj:`str`: scheme value
        """
        return self.parsed.scheme

    @property
    def url(self):
        """Get scheme, hostname, and port from :attr:`ParserUrl.parsed`.

        Returns:
            :obj:`str`: schema, hostname, and port unparsed values
        """
        return self.unparse_base(parsed_result=self.parsed)

    @property
    def url_full(self):
        """Get full URL from :attr:`ParserUrl.parsed`.

        Returns:
            :obj:`str`: full unparsed url
        """
        return self.unparse_all(parsed_result=self.parsed)

    @property
    def parsed_str(self):
        """Get a str value of :attr:`ParserUrl.parsed`.

        Returns:
            :obj:`str`: str value of :attr:`ParserUrl.parsed`
        """
        parsed = getattr(self, "parsed", None)
        attrs = [
            "scheme",
            "netloc",
            "hostname",
            "port",
            "path",
            "params",
            "query",
            "fragment",
        ]
        atmpl = "{a}={v!r}".format
        attrs = [atmpl(a=a, v="{}".format(getattr(parsed, a, "")) or "") for a in attrs]
        return ", ".join(attrs)

    def make_netloc(self, host, port):
        """Create netloc from host and port.

        Args:
            host (:obj:`str`): host part to use in netloc
            port (:obj:`str`): port part to use in netloc

        Returns:
            :obj:`str`: host and port values joined by :

        """
        return ":".join([x for x in [host, port] if x])

    def reparse(self, parsed, default_scheme=""):
        """Reparse a parsed URL into a parsed URL with values fixed.

        Args:
            parsed (:obj:`urllib.parse.ParseResult`): parsed URL to reparse
            default_scheme (:obj:`str`, optional): default ``""`` -
                default scheme to use if URL does not contain a scheme

        Returns:
            :obj:`urllib.parse.ParseResult`: reparsed result
        """
        scheme, netloc, path, params, query, fragment = parsed
        host = parsed.hostname
        port = format(parsed.port or "")

        if not netloc and scheme and path and path.split("/")[0].isdigit():
            """For case:
            >>> urllib.parse.urlparse('host:443/')
            ParseResult(
                scheme='host', netloc='', path='443/', params='', query='', fragment=''
            )
            """
            host = scheme  # switch host from scheme to host
            port = path.split("/")[0]  # remove / from path and assign to port
            path = ""  # empty out path
            scheme = default_scheme
            netloc = ":".join([host, port])

        if not netloc and path:
            """For cases:
            >>> urllib.parse.urlparse('host:443')
            ParseResult(
                scheme='', netloc='', path='host:443', params='', query='', fragment=''
            )
            >>> urllib.parse.urlparse('host')
            ParseResult(
                scheme='', netloc='', path='host', params='', query='', fragment=''
            )
            """
            netloc, path = path, netloc
            if ":" in netloc:  # pragma: no cover
                # can't get this to trigger anymore, ignore test coverage
                host, port = netloc.split(":", 1)
                netloc = ":".join([host, port]) if port else host
            else:
                host = netloc

        scheme = scheme or default_scheme
        if not scheme and port:
            if format(port) == "443":
                scheme = "https"
            elif format(port) == "80":
                scheme = "http"

        if not port:
            if scheme == "https":
                netloc = self.make_netloc(host, "443")
            elif scheme == "http":
                netloc = self.make_netloc(host, "80")

        pass2 = urlunparse((scheme, netloc, path, params, query, fragment))
        return urlparse(pass2)

    def unparse_base(self, parsed_result):
        """Unparse a parsed URL into just the scheme, hostname, and port parts.

        Args:
            parsed (:obj:`urllib.parse.ParseResult`): parsed URL to unparse

        Returns:
            :obj:`str`: unparsed url
        """
        # only unparse self.parsed into url with scheme and netloc
        bits = (parsed_result.scheme, parsed_result.netloc, "", "", "", "")
        return urlunparse(bits)

    def unparse_all(self, parsed_result):
        """Unparse a parsed URL with all the parts.

        Args:
            parsed (:obj:`urllib.parse.ParseResult`): parsed URL to unparse

        Returns:
            :obj:`str`: unparsed url
        """
        return urlunparse(parsed_result)



# -*- coding: utf-8 -*-
"""Command line interface for Axonius API Client."""
# from ....constants import DEFAULT_PERM, PERM_SETS, VALID_PERMS
from ....tools import dt_parse, json_dump, listify
from ...context import click

EXPORT = click.option(
    "--export-format",
    "-xf",
    "export_format",
    type=click.Choice(["json", "str"]),
    help="Format of to export data in",
    default="str",
    show_envvar=True,
    show_default=True,
)


def handle_export(ctx, data, export_format, **kwargs):
    """Get roles."""
    if export_format == "json":
        click.secho(json_dump(data))
        ctx.exit(0)

    if export_format == "str":
        for item in listify(data):
            first_name = item["first_name"]
            last_name = item["last_name"]
            role = item["role_obj"]
            role_name = role["name"]
            source = item["source"]
            user_name = item["user_name"]
            last_login = item.get("last_login")
            if last_login:
                last_login = str(dt_parse(last_login))

            lines = [
                f"Name: {user_name!r}",
                f"Role: {role_name!r}",
                f"Last Login: {last_login!r}",
                f"First Name: {first_name!r}",
                f"Last Name: {last_name!r}",
                f"Source: {source!r}",
            ]
            click.secho(", ".join(lines))
        ctx.exit(0)

    ctx.exit(1)



# -*- coding: utf-8 -*-
"""Command line interface for Axonius API Client."""
from ...tools import json_dump
from ..context import CONTEXT_SETTINGS, click
from ..options import AUTH, NODE, add_options

OPTIONS = [
    *AUTH,
    *NODE,
    click.option(
        "--input-file",
        "-if",
        "input_file",
        help="File to upload (from path or piped via STDIN)",
        default="-",
        type=click.File(mode="r"),
        show_envvar=True,
        show_default=True,
    ),
]


@click.command(name="file-upload", context_settings=CONTEXT_SETTINGS)
@add_options(OPTIONS)
@click.pass_context
def cmd(ctx, url, key, secret, input_file, **kwargs):
    """Upload a file to an adapter for later use with a connection."""
    client = ctx.obj.start_client(url=url, key=key, secret=secret)
    stream_name = format(getattr(input_file, "name", input_file))
    kwargs["file_name"] = stream_name
    kwargs["field_name"] = stream_name
    kwargs["file_content"] = ctx.obj.read_stream(stream=input_file)

    with ctx.obj.exc_wrap(wraperror=ctx.obj.wraperror):
        rows = client.adapters.file_upload(**kwargs)

    ctx.obj.echo_ok(f"File uploaded")
    click.echo(json_dump(rows, indent=None))



# -*- coding: utf-8 -*-
"""Test suite for assets."""
import copy
import io
import logging
import sys

import pytest
from axonius_api_client.api.asset_callbacks import get_callbacks_cls
from axonius_api_client.constants import (AGG_ADAPTER_NAME, AGG_ADAPTER_TITLE,
                                          FIELD_TRIM_LEN, SCHEMAS_CUSTOM)
from axonius_api_client.exceptions import ApiError, NotFoundError

from ...meta import TAGS
from ...utils import log_check


def load_test_data(apiobj):
    """Pass."""
    apiobj.TEST_DATA = getattr(apiobj, "TEST_DATA", {})

    if not apiobj.TEST_DATA.get("fields_map"):
        apiobj.TEST_DATA["fields_map"] = fields_map = apiobj.fields.get()

    if not apiobj.TEST_DATA.get("assets"):
        apiobj.TEST_DATA["assets"] = apiobj.get(max_rows=2000, fields_map=fields_map)

    if not apiobj.TEST_DATA.get("cb_assets"):
        field_complex = apiobj.TEST_DATA["field_complex"]
        try:
            apiobj.fields.get_field_name(value=field_complex)
        except NotFoundError:
            # XXX users no longer seem to have any complex field data as of 04/23/2020
            apiobj.TEST_DATA["field_complex"] = None
            apiobj.TEST_DATA["field_complexes"] = apiobj.fields_default
            apiobj.TEST_DATA["cb_assets_query"] = None

        field_complexes = apiobj.TEST_DATA["field_complexes"]
        cb_assets_query = apiobj.TEST_DATA["cb_assets_query"]
        apiobj.TEST_DATA["cb_assets"] = apiobj.get(
            query=cb_assets_query,
            max_rows=2000,
            fields=field_complexes,
            fields_map=fields_map,
        )

    return apiobj


class Callbacks:
    """Pass."""

    def get_cbobj(self, apiobj, cbexport, getargs=None):
        """Pass."""
        state = {
            "max_pages": None,
            "max_rows": None,
            "page": {},
            "page_cursor": None,
            "page_left": 0,
            "page_num": 1,
            "page_size": 2000,
            "page_sleep": 0,
            "page_total": 0,
            "rows_fetched": 0,
            "rows_page": 0,
            "rows_processed": 0,
            "rows_start": 0,
            "rows_total": 0,
            "time_fetch": 0,
            "time_page": 0,
            "use_cursor": False,
        }
        getargs = getargs or {}
        query = apiobj.TEST_DATA["cb_assets_query"]
        fields = apiobj.TEST_DATA["field_complexes"]
        cbcls = get_callbacks_cls(export=cbexport)
        assert cbcls.CB_NAME == cbexport

        cbobj = cbcls(
            apiobj=apiobj,
            fields_map=apiobj.TEST_DATA["fields_map"],
            getargs=getargs,
            fields=fields,
            query=query,
            state=state,
        )
        assert cbobj.CB_NAME == cbexport
        assert cbobj.APIOBJ == apiobj
        assert cbobj.FIELDS_MAP == apiobj.TEST_DATA["fields_map"]
        assert cbobj.FIELDS == fields
        assert cbobj.QUERY == query
        assert cbobj.GETARGS == getargs
        assert cbobj.STATE == state

        assert isinstance(cbobj.args_map, list)
        assert isinstance(cbobj.args_strs, list)

        assert isinstance(cbobj.TAG_ROWS_ADD, list) and not cbobj.TAG_ROWS_ADD
        assert isinstance(cbobj.TAG_ROWS_REMOVE, list) and not cbobj.TAG_ROWS_REMOVE

        assert isinstance(cbobj.LOG, logging.Logger)
        assert isinstance(cbobj.RAN, list)
        assert not cbobj.RAN

        assert cbobj.__class__.__name__ in str(cbobj)
        assert cbobj.__class__.__name__ in repr(cbobj)

        assert isinstance(cbobj.SCHEMAS_ALL, list)
        assert cbobj.SCHEMAS_ALL

        assert isinstance(cbobj.adapter_map, dict)
        assert cbobj.adapter_map

        assert isinstance(cbobj.SCHEMAS_FIELDS, list)
        assert cbobj.SCHEMAS_FIELDS
        for schema in cbobj.SCHEMAS_FIELDS:
            assert schema in cbobj.SCHEMAS_ALL

        return cbobj

    def test_start_stop(self, cbexport, apiobj, caplog):
        """Pass."""
        getargs = {}
        cbobj = self.get_cbobj(apiobj=apiobj, cbexport=cbexport, getargs=getargs)
        assert not caplog.records

        cbobj.start()
        log_check(caplog=caplog, entries=["Starting"], exists=True)

        cbobj.stop()
        log_check(caplog=caplog, entries=["Stopping"], exists=True)

    def test_report_adapters_missing_false(self, cbexport, apiobj):
        """Pass."""
        row = copy.deepcopy(apiobj.TEST_DATA["cb_assets"][0])

        report_schemas = SCHEMAS_CUSTOM["report_adapters_missing"]
        getargs = {"report_adapters_missing": False}
        cbobj = self.get_cbobj(apiobj=apiobj, cbexport=cbexport, getargs=getargs)

        row_ret = cbobj.report_adapters_missing(row=copy.deepcopy(row))
        assert isinstance(row_ret, dict)
        assert row == row_ret

        assert isinstance(cbobj.SCHEMAS_CUSTOM, list)
        for field, schema in report_schemas.items():
            assert schema not in cbobj.SCHEMAS_CUSTOM

        assert not cbobj.SCHEMAS_CUSTOM

    def test_report_adapters_missing_true(self, cbexport, apiobj):
        """Pass."""
        row = copy.deepcopy(apiobj.TEST_DATA["cb_assets"][0])

        report_schemas = SCHEMAS_CUSTOM["report_adapters_missing"]
        getargs = {"report_adapters_missing": True}
        cbobj = self.get_cbobj(apiobj=apiobj, cbexport=cbexport, getargs=getargs)

        row_ret = cbobj.report_adapters_missing(row=copy.deepcopy(row))
        assert isinstance(row_ret, dict)
        assert row != row_ret

        assert isinstance(cbobj.SCHEMAS_CUSTOM, list)
        for field, schema in report_schemas.items():
            assert schema["name_qual"] in row_ret
            assert schema in cbobj.SCHEMAS_CUSTOM
            assert schema in cbobj.schemas_final(flat=False)
            assert schema in cbobj.schemas_final(flat=True)

    def test_firstpage(self, cbexport, apiobj, caplog):
        """Pass."""
        getargs = {"first_page": True}
        cbobj = self.get_cbobj(apiobj=apiobj, cbexport=cbexport, getargs=getargs)
        log_entries = ["First page "]

        cbobj.first_page()
        assert "first_page" in cbobj.RAN
        log_check(caplog=caplog, entries=log_entries, exists=True)

        caplog.clear()
        cbobj.first_page()
        log_check(caplog=caplog, entries=log_entries, exists=False)

    def test_field_null_true(self, cbexport, apiobj):
        """Pass."""
        row = copy.deepcopy(apiobj.TEST_DATA["cb_assets"][0])

        field_complex = apiobj.TEST_DATA["field_complex"]
        if field_complex is None:
            pytest.skip(f"No complex field found for {apiobj}")
        field_complexes = apiobj.TEST_DATA["field_complexes"]

        schema = apiobj.fields.get_field_schema(
            value=field_complex,
            schemas=apiobj.TEST_DATA["fields_map"][AGG_ADAPTER_NAME],
        )

        getargs = {"field_null": True}
        cbobj = self.get_cbobj(apiobj=apiobj, cbexport=cbexport, getargs=getargs)

        row_ret = cbobj.field_null(row=copy.deepcopy(row))
        assert isinstance(row_ret, dict)
        assert row != row_ret

        for field in field_complexes:
            assert field in row_ret

        for sub_field in schema["sub_fields"]:
            for sub_value in row_ret[field_complex]:
                if sub_field["is_root"]:
                    assert sub_field["name"] in sub_value
                else:
                    assert sub_field["name"] not in sub_value

    def test_field_null_false(self, cbexport, apiobj):
        """Pass."""
        row = copy.deepcopy(apiobj.TEST_DATA["cb_assets"][0])
        getargs = {"field_null": False}
        cbobj = self.get_cbobj(apiobj=apiobj, cbexport=cbexport, getargs=getargs)

        row_ret = cbobj.field_null(row=copy.deepcopy(row))
        assert isinstance(row_ret, dict)
        assert row == row_ret

    def test_field_null_exclude(self, cbexport, apiobj):
        """Pass."""
        row = copy.deepcopy(apiobj.TEST_DATA["cb_assets"][0])
        row.pop("adapters", None)
        getargs = {"field_null": True, "field_excludes": ["adapters"]}
        cbobj = self.get_cbobj(apiobj=apiobj, cbexport=cbexport, getargs=getargs)

        row_ret = cbobj.field_null(row=copy.deepcopy(row))
        assert isinstance(row_ret, dict)
        assert "adapters" not in row_ret

    def test_tags_add(self, cbexport, apiobj, caplog):
        """Pass."""
        row = copy.deepcopy(apiobj.TEST_DATA["cb_assets"][0])
        row_id = row["internal_axon_id"]

        getargs = {"tags_add": TAGS}
        cbobj = self.get_cbobj(apiobj=apiobj, cbexport=cbexport, getargs=getargs)

        row_ret = cbobj.tags_add(row=copy.deepcopy(row))
        assert row == row_ret
        assert {"internal_axon_id": row_id} in cbobj.TAG_ROWS_ADD

        cbobj.do_tagging()
        log_entries = ["tags.*assets"]
        log_check(caplog=caplog, entries=log_entries, exists=True)

        all_tags = apiobj.labels.get()
        row_refetch = apiobj.get_by_value(
            field="internal_axon_id", value=row_id, field_manual=True
        )[0]

        row_tags = row_refetch.get(apiobj.FIELD_TAGS, [])
        for tag in TAGS:
            assert tag in all_tags
            assert tag in row_tags

    def test_tags_remove(self, cbexport, apiobj, caplog):
        """Pass."""
        row = copy.deepcopy(apiobj.TEST_DATA["cb_assets"][0])
        row_id = row["internal_axon_id"]

        getargs = {"tags_remove": TAGS}
        cbobj = self.get_cbobj(apiobj=apiobj, cbexport=cbexport, getargs=getargs)

        row_ret = cbobj.tags_remove(row=copy.deepcopy(row))
        assert row == row_ret
        assert {"internal_axon_id": row_id} in cbobj.TAG_ROWS_REMOVE

        cbobj.do_tagging()

        log_entries = ["tags.*assets"]
        log_check(caplog=caplog, entries=log_entries, exists=True)

        row_refetch = apiobj.get_by_value(
            field="internal_axon_id", value=row_id, field_manual=True
        )[0]

        row_tags = row_refetch.get(apiobj.FIELD_TAGS, [])
        for tag in TAGS:
            assert tag not in row_tags

    def test_tags_add_empty(self, cbexport, apiobj, caplog):
        """Pass."""
        row = copy.deepcopy(apiobj.TEST_DATA["cb_assets"][0])

        getargs = {"tags_add": []}
        cbobj = self.get_cbobj(apiobj=apiobj, cbexport=cbexport, getargs=getargs)

        row_ret = cbobj.tags_add(row=copy.deepcopy(row))
        assert row == row_ret
        assert not cbobj.TAG_ROWS_ADD

        cbobj.do_tagging()

        log_entries = ["tags.*assets"]
        log_check(caplog=caplog, entries=log_entries, exists=False)

    def test_tags_remove_empty(self, cbexport, apiobj, caplog):
        """Pass."""
        row = copy.deepcopy(apiobj.TEST_DATA["cb_assets"][0])

        getargs = {"tags_remove": []}
        cbobj = self.get_cbobj(apiobj=apiobj, cbexport=cbexport, getargs=getargs)

        row_ret = cbobj.tags_remove(row=copy.deepcopy(row))
        assert row == row_ret
        assert not cbobj.TAG_ROWS_REMOVE

        cbobj.do_tagging()

        log_entries = ["tags.*assets"]
        log_check(caplog=caplog, entries=log_entries, exists=False)

    def test_field_excludes_empty(self, cbexport, apiobj):
        """Pass."""
        row = copy.deepcopy(apiobj.TEST_DATA["cb_assets"][0])

        getargs = {"field_excludes": []}
        cbobj = self.get_cbobj(apiobj=apiobj, cbexport=cbexport, getargs=getargs)

        row_ret = cbobj.field_excludes(row=copy.deepcopy(row))
        assert row == row_ret

    def test_field_excludes(self, cbexport, apiobj):
        """Pass."""
        row = copy.deepcopy(apiobj.TEST_DATA["cb_assets"][0])

        fields = ["internal_axon_id", "adapters", "adapters_list_length"]

        getargs = {"field_excludes": fields}
        cbobj = self.get_cbobj(apiobj=apiobj, cbexport=cbexport, getargs=getargs)

        row_ret = cbobj.field_excludes(row=copy.deepcopy(row))
        assert row != row_ret

        for field in fields:
            assert field not in row_ret

    def test_field_excludes_sub(self, cbexport, apiobj):
        """Pass."""
        row = copy.deepcopy(apiobj.TEST_DATA["cb_assets"][0])

        field_complex = apiobj.TEST_DATA["field_complex"]
        if field_complex is None:
            pytest.skip(f"No complex field found for {apiobj}")

        sub_exclude = list(row[field_complex][0])[0]
        fields = ["internal_axon_id", sub_exclude]

        getargs = {"field_excludes": fields}
        cbobj = self.get_cbobj(apiobj=apiobj, cbexport=cbexport, getargs=getargs)

        row_ret = cbobj.field_excludes(row=copy.deepcopy(row))
        assert row != row_ret

        for field in fields:
            assert field not in row_ret
        for item in row_ret[field_complex]:
            assert sub_exclude not in item

    def test_field_join_true(self, cbexport, apiobj):
        """Pass."""
        row = copy.deepcopy(apiobj.TEST_DATA["cb_assets"][0])

        getargs = {"field_join": True}
        cbobj = self.get_cbobj(apiobj=apiobj, cbexport=cbexport, getargs=getargs)

        row_ret = cbobj.field_join(row=copy.deepcopy(row))

        assert row != row_ret
        for field in row:
            if isinstance(row[field], list):
                assert isinstance(row_ret[field], str)

    def test_field_join_false(self, cbexport, apiobj):
        """Pass."""
        row = copy.deepcopy(apiobj.TEST_DATA["cb_assets"][0])

        getargs = {"field_join": False}
        cbobj = self.get_cbobj(apiobj=apiobj, cbexport=cbexport, getargs=getargs)

        row_ret = cbobj.field_join(row=copy.deepcopy(row))
        assert row == row_ret

    def test_field_join_trim(self, cbexport, apiobj):
        """Pass."""
        row = copy.deepcopy(apiobj.TEST_DATA["cb_assets"][0])
        row["test"] = ("aaaa " * (FIELD_TRIM_LEN + 1000)).split()

        getargs = {"field_join": True}
        cbobj = self.get_cbobj(apiobj=apiobj, cbexport=cbexport, getargs=getargs)

        row_ret = cbobj.field_join(row=copy.deepcopy(row))
        assert row != row_ret
        assert "TRIMMED" in row_ret["test"]

        exp_len = FIELD_TRIM_LEN + 100
        test_len = len(row_ret["test"])
        assert test_len <= exp_len

    def test_field_join_trim_disabled(self, cbexport, apiobj):
        """Pass."""
        row = copy.deepcopy(apiobj.TEST_DATA["cb_assets"][0])
        row["test"] = ("aaaa " * (FIELD_TRIM_LEN + 1000)).split()

        getargs = {"field_join": True, "field_join_trim": 0}
        cbobj = self.get_cbobj(apiobj=apiobj, cbexport=cbexport, getargs=getargs)

        row_ret = cbobj.field_join(row=copy.deepcopy(row))
        assert row != row_ret
        assert "TRIMMED" not in row_ret["test"]

    def test_field_join_trim_custom_joiner(self, cbexport, apiobj):
        """Pass."""
        row = copy.deepcopy(apiobj.TEST_DATA["cb_assets"][0])
        row["test"] = ("aaaa " * (FIELD_TRIM_LEN + 1000)).split()

        getargs = {"field_join": True, "field_join_value": "!!"}
        cbobj = self.get_cbobj(apiobj=apiobj, cbexport=cbexport, getargs=getargs)

        row_ret = cbobj.field_join(row=copy.deepcopy(row))
        assert row != row_ret
        assert "!!" in row_ret["test"]

    def test_field_titles_true(self, cbexport, apiobj):
        """Pass."""
        row = copy.deepcopy(apiobj.TEST_DATA["cb_assets"][0])
        row_titles = list(row)

        getargs = {"field_titles": True}
        cbobj = self.get_cbobj(apiobj=apiobj, cbexport=cbexport, getargs=getargs)

        row_ret = cbobj.field_titles(row=copy.deepcopy(row))
        assert row != row_ret

        row_ret_titles = list(row_ret)
        assert row_titles != row_ret_titles
        for x in row_ret_titles:
            assert f"{AGG_ADAPTER_TITLE}: " in x
            assert "specific_data.data." not in x

    def test_field_titles_false(self, cbexport, apiobj):
        """Pass."""
        row = copy.deepcopy(apiobj.TEST_DATA["cb_assets"][0])

        getargs = {"field_titles": False}
        cbobj = self.get_cbobj(apiobj=apiobj, cbexport=cbexport, getargs=getargs)

        row_ret = cbobj.field_titles(row=copy.deepcopy(row))
        assert row == row_ret

    def test_field_flatten_true(self, cbexport, apiobj):
        """Pass."""
        row = copy.deepcopy(apiobj.TEST_DATA["cb_assets"][0])
        field_complex = apiobj.TEST_DATA["field_complex"]
        if field_complex is None:
            pytest.skip(f"No complex field found for {apiobj}")
        schema = apiobj.fields.get_field_schema(
            value=field_complex,
            schemas=apiobj.TEST_DATA["fields_map"][AGG_ADAPTER_NAME],
        )
        sub_fields = schema["sub_fields"]
        sub_quals = [x["name_qual"] for x in sub_fields if x["is_root"]]

        getargs = {"field_flatten": True}
        cbobj = self.get_cbobj(apiobj=apiobj, cbexport=cbexport, getargs=getargs)

        row_ret = cbobj.field_flatten(row=copy.deepcopy(row))
        assert row != row_ret

        assert any([y for x in sub_quals for y in row_ret[x]])
        assert schema["name_qual"] not in row_ret

        row_sub_fields = set()

        for field, value in row.items():
            if not isinstance(value, list) or not value:
                continue

            if not isinstance(value[0], dict):
                continue

            # make sure the orginal complex field got removed from the new row
            assert field not in row_ret

            for sub_value in value:
                row_sub_fields.update(list(sub_value))

        for field, value in row_ret.items():
            assert not isinstance(value, dict)

            if not isinstance(value, list):
                continue

            # assert no list items are complex
            assert not any([isinstance(i, (list, dict)) for i in value])

            if field not in row:
                # assert subfields from complex added to new row are fully qual'd
                assert "specific_data.data." in field

            if field in row:
                continue

            if not any([field.endswith(x) for x in row_sub_fields]):
                assert all([x is None for x in value])

    def test_field_flatten_exclude_sub(self, cbexport, apiobj):
        """Pass."""
        row = copy.deepcopy(apiobj.TEST_DATA["cb_assets"][0])
        field_complex = apiobj.TEST_DATA["field_complex"]
        if field_complex is None:
            pytest.skip(f"No complex field found for {apiobj}")

        schema = apiobj.fields.get_field_schema(
            value=field_complex,
            schemas=apiobj.TEST_DATA["fields_map"][AGG_ADAPTER_NAME],
        )
        sub_field = schema["sub_fields"][0]
        sub_field_name = sub_field["name"]
        sub_field_name_qual = sub_field["name_qual"]

        getargs = {
            "field_flatten": True,
            "field_excludes": [sub_field_name, "internal_axon_id"],
        }
        cbobj = self.get_cbobj(apiobj=apiobj, cbexport=cbexport, getargs=getargs)

        row_ret = cbobj.field_flatten(row=copy.deepcopy(row))
        assert row != row_ret
        assert sub_field_name_qual not in row_ret

    def test_field_flatten_false(self, cbexport, apiobj):
        """Pass."""
        row = copy.deepcopy(apiobj.TEST_DATA["cb_assets"][0])

        getargs = {"field_flatten": False}
        cbobj = self.get_cbobj(apiobj=apiobj, cbexport=cbexport, getargs=getargs)

        row_ret = cbobj.field_flatten(row=copy.deepcopy(row))
        assert row == row_ret

    def test_field_flatten_custom_null(self, cbexport, apiobj):
        """Pass."""
        if apiobj.TEST_DATA["field_complex"] is None:
            pytest.skip(f"No complex field found for {apiobj}")

        row = copy.deepcopy(apiobj.TEST_DATA["cb_assets"][0])

        getargs = {"field_flatten": True, "field_null_value": "badwolf"}
        cbobj = self.get_cbobj(apiobj=apiobj, cbexport=cbexport, getargs=getargs)

        row_ret = cbobj.field_flatten(row=copy.deepcopy(row))

        row_sub_fields = set()
        assert row != row_ret

        for field, value in row.items():
            if not isinstance(value, list) or not value:
                continue

            if not isinstance(value[0], dict):
                continue

            # make sure the orginal complex field got removed from the new row
            assert field not in row_ret

            for sub_value in value:
                row_sub_fields.update(list(sub_value))

        for field, value in row_ret.items():
            assert not isinstance(value, dict)

            if not isinstance(value, list):
                continue

            # assert no list items are complex
            assert not any([isinstance(i, (list, dict)) for i in value])

            if field not in row:
                # assert subfields from complex added to new row are fully qual'd
                assert "specific_data.data." in field

            if field in row:
                continue

            if not any([field.endswith(x) for x in row_sub_fields]):
                assert all([x == "badwolf" for x in value])

    def test_field_explode_complex(self, cbexport, apiobj):
        """Pass."""
        row = copy.deepcopy(apiobj.TEST_DATA["cb_assets"][0])
        field_complex = apiobj.TEST_DATA["field_complex"]
        if field_complex is None:
            pytest.skip(f"No complex field found for {apiobj}")

        getargs = {"field_explode": field_complex}
        cbobj = self.get_cbobj(apiobj=apiobj, cbexport=cbexport, getargs=getargs)

        row_ret = cbobj.field_explode(row=copy.deepcopy(row))
        assert isinstance(row_ret, list)
        assert row != row_ret
        assert len(row) > len(row_ret)
        for row_x in row_ret:
            assert field_complex not in row_x

    def test_field_explode_simple(self, cbexport, apiobj):
        """Pass."""
        key = "adapters"
        row = copy.deepcopy(apiobj.TEST_DATA["cb_assets"][0])
        row_val = row["adapters"]
        row_len = len(row_val)
        getargs = {"field_explode": key}
        cbobj = self.get_cbobj(apiobj=apiobj, cbexport=cbexport, getargs=getargs)

        row_ret = cbobj.field_explode(row=copy.deepcopy(row))
        assert isinstance(row_ret, list)
        assert row != row_ret
        assert len(row_ret) == row_len
        for idx, row_x in enumerate(row_ret):
            value = row_x[key]
            assert isinstance(value, str)
            assert value == row_val[idx]

    def test_field_explode_exclude(self, cbexport, apiobj):
        """Pass."""
        row = copy.deepcopy(apiobj.TEST_DATA["cb_assets"][0])

        getargs = {"field_explode": "adapters", "field_excludes": ["adapters"]}
        cbobj = self.get_cbobj(apiobj=apiobj, cbexport=cbexport, getargs=getargs)

        row_ret = cbobj.field_explode(row=copy.deepcopy(row))
        assert isinstance(row_ret, list)
        assert len(row_ret) == 1
        assert row == row_ret[0]

    def test_field_explode_none(self, cbexport, apiobj):
        """Pass."""
        row = copy.deepcopy(apiobj.TEST_DATA["cb_assets"][0])

        getargs = {}
        cbobj = self.get_cbobj(apiobj=apiobj, cbexport=cbexport, getargs=getargs)

        row_ret = cbobj.field_explode(row=copy.deepcopy(row))
        assert isinstance(row_ret, list)
        assert len(row_ret) == 1
        assert row == row_ret[0]

    def test_field_explode_schema_error(self, cbexport, apiobj):
        """Pass."""
        getargs = {"field_explode": "badwolf"}
        cbobj = self.get_cbobj(apiobj=apiobj, cbexport=cbexport, getargs=getargs)
        with pytest.raises(ApiError):
            cbobj.explode_schema

    def test_fd_stdout_open_no_close(self, cbexport, apiobj):
        """Pass."""
        getargs = {}
        cbobj = self.get_cbobj(apiobj=apiobj, cbexport=cbexport, getargs=getargs)

        fd = cbobj.open_fd()

        assert fd == sys.stdout
        assert cbobj._fd == sys.stdout
        assert not cbobj._fd_close

        cbobj.close_fd()
        sys.stdout.write("should still work")

    def test_fd_custom_open_close_false(self, cbexport, apiobj):
        """Pass."""
        io_fd = io.StringIO()

        getargs = {"export_fd": io_fd, "export_fd_close": False}
        cbobj = self.get_cbobj(apiobj=apiobj, cbexport=cbexport, getargs=getargs)

        fd = cbobj.open_fd()

        assert fd == io_fd
        assert cbobj._fd == io_fd
        assert not cbobj._fd_close

        cbobj.close_fd()

        value = fd.getvalue()
        assert value == "\n"
        fd.write("")

    def test_fd_custom_open_close_true(self, cbexport, apiobj):
        """Pass."""
        io_fd = io.StringIO()

        getargs = {"export_fd": io_fd, "export_fd_close": True}
        cbobj = self.get_cbobj(apiobj=apiobj, cbexport=cbexport, getargs=getargs)

        fd = cbobj.open_fd()
        assert fd == io_fd
        assert cbobj._fd == io_fd
        assert cbobj._fd_close

        cbobj.close_fd()

        with pytest.raises(ValueError):
            fd.getvalue()

        with pytest.raises(ValueError):
            fd.write("")

    def test_fd_path_open_close_true(self, cbexport, apiobj, tmp_path):
        """Pass."""
        export_file = tmp_path / "badwolf.txt"
        getargs = {"export_file": export_file}
        cbobj = self.get_cbobj(apiobj=apiobj, cbexport=cbexport, getargs=getargs)

        cbobj.open_fd()

        assert cbobj._file_path.name == export_file.name
        assert cbobj._file_mode == "created"
        assert cbobj._fd_close

        cbobj.close_fd()
        assert export_file.is_file()
        assert export_file.read_text() == "\n"

        with pytest.raises(ValueError):
            cbobj._fd.write("")

    def test_fd_path_open_close_false(self, cbexport, apiobj, tmp_path):
        """Pass."""
        export_file = tmp_path / "badwolf.txt"
        getargs = {"export_file": export_file, "export_fd_close": False}
        cbobj = self.get_cbobj(apiobj=apiobj, cbexport=cbexport, getargs=getargs)

        cbobj.open_fd()

        assert cbobj._file_path.name == export_file.name
        assert cbobj._file_mode == "created"
        assert not cbobj._fd_close

        cbobj.close_fd()
        assert export_file.is_file()
        cbobj._fd.write(" ")
        cbobj._fd.close()
        assert export_file.read_text() == "\n "

    def test_fd_path_overwrite_true(self, cbexport, apiobj, tmp_path):
        """Pass."""
        export_file = tmp_path / "badwolf.txt"
        export_file.touch()
        getargs = {"export_file": export_file, "export_overwrite": True}
        cbobj = self.get_cbobj(apiobj=apiobj, cbexport=cbexport, getargs=getargs)

        cbobj.open_fd()

        assert cbobj._file_path.name == export_file.name
        assert cbobj._file_mode == "overwrote"
        assert cbobj._fd_close

        cbobj.close_fd()
        assert export_file.is_file()

    def test_fd_path_overwrite_false(self, cbexport, apiobj, tmp_path):
        """Pass."""
        export_file = tmp_path / "badwolf.txt"
        export_file.touch()
        getargs = {"export_file": export_file, "export_overwrite": False}
        cbobj = self.get_cbobj(apiobj=apiobj, cbexport=cbexport, getargs=getargs)

        with pytest.raises(ApiError):
            cbobj.open_fd()

    def test_echo_ok_doecho_yes(self, cbexport, apiobj, capsys, caplog):
        """Pass."""
        entry = "xxxxxxx"

        getargs = {"do_echo": True}
        cbobj = self.get_cbobj(apiobj=apiobj, cbexport=cbexport, getargs=getargs)
        cbobj.echo(msg=entry)
        capture = capsys.readouterr()
        assert f"{entry}\n" in capture.err
        assert not capture.out
        log_check(caplog=caplog, entries=[entry], exists=False)

    def test_echo_error_doecho_yes(self, cbexport, apiobj, capsys, caplog):
        """Pass."""
        entry = "xxxxxxx"

        getargs = {"do_echo": True}
        cbobj = self.get_cbobj(apiobj=apiobj, cbexport=cbexport, getargs=getargs)
        with pytest.raises(SystemExit):
            cbobj.echo(msg=entry, error=ApiError)

        capture = capsys.readouterr()
        assert f"{entry}\n" in capture.err
        assert not capture.out
        log_check(caplog=caplog, entries=[entry], exists=True)

    def test_echo_ok_doecho_no(self, cbexport, apiobj, capsys, caplog):
        """Pass."""
        entry = "xxxxxxx"

        getargs = {"do_echo": False}
        cbobj = self.get_cbobj(apiobj=apiobj, cbexport=cbexport, getargs=getargs)
        cbobj.echo(msg=entry)
        capture = capsys.readouterr()
        assert not capture.err
        assert not capture.out
        log_check(caplog=caplog, entries=[entry], exists=True)

    def test_echo_error_doecho_no(self, cbexport, apiobj, capsys, caplog):
        """Pass."""
        entry = "xxxxxxx"

        getargs = {"do_echo": False}
        cbobj = self.get_cbobj(apiobj=apiobj, cbexport=cbexport, getargs=getargs)
        with pytest.raises(ApiError):
            cbobj.echo(msg=entry, error=ApiError)

        capture = capsys.readouterr()
        assert not capture.err
        assert not capture.out
        log_check(caplog=caplog, entries=[entry], exists=True)

    def test_get_callbacks_cls_error(self):
        """Pass."""
        with pytest.raises(ApiError):
            get_callbacks_cls(export="badwolf")



# -*- coding: utf-8 -*-
"""API models for working with adapters and connections."""
import time

from ...constants import CNX_GONE, CNX_RETRY, CNX_SANE_DEFAULTS, DEFAULT_NODE
from ...exceptions import (CnxAddError, CnxGoneError, CnxTestError,
                           CnxUpdateError, ConfigInvalidValue, ConfigRequired,
                           NotFoundError)
from ...tools import json_load, pathlib
from ..mixins import ChildMixins
from ..parsers.config import (config_build, config_default, config_empty,
                              config_info, config_required, config_unchanged,
                              config_unknown)
from ..parsers.tables import tablize_cnxs, tablize_schemas


class Cnx(ChildMixins):
    """API model for working with adapter connections."""

    def add(self, adapter_name, adapter_node=DEFAULT_NODE, **kwargs):
        """Pass."""
        kwargs_config = kwargs.pop("kwargs_config", {})
        kwargs.update(kwargs_config)

        adapter = self.parent.get_by_name(name=adapter_name, node=adapter_node)
        cnx_schemas = adapter["schemas"]["cnx"]
        adapter_name = adapter["name"]
        adapter_node_name = adapter["node_name"]
        adapter_name_raw = adapter["name_raw"]
        adapter_node_id = adapter["node_id"]

        source = f"adding connection for adapter {adapter_name!r}"

        new_config = self.build_config(
            cnx_schemas=cnx_schemas,
            old_config={},
            new_config=kwargs,
            source=source,
            adapter_name=adapter_name,
            adapter_node=adapter_node_name,
        )

        sane_defaults = self.get_sane_defaults(adapter_name=adapter_name)

        config_default(
            schemas=cnx_schemas,
            new_config=new_config,
            source=source,
            sane_defaults=sane_defaults,
        )

        config_empty(schemas=cnx_schemas, new_config=new_config, source=source)

        config_required(schemas=cnx_schemas, new_config=new_config, source=source)

        result = self._add(
            adapter_name_raw=adapter_name_raw,
            adapter_node_id=adapter_node_id,
            new_config=new_config,
        )

        error_in_status = result.get("status", "") == "error"
        error_empty = bool(result.get("error", ""))

        cnx_new = self.get_by_uuid(
            cnx_uuid=result["id"],
            adapter_name=adapter_name,
            adapter_node=adapter_node,
            retry=CNX_RETRY,
        )

        if any([error_in_status, error_empty]):
            rkw = ["{}: {}".format(k, v) for k, v in result.items()]
            rkw = "\n  " + "\n  ".join(rkw)

            err = f"Connection was added but had a failure connecting:{rkw}"
            exc = CnxAddError(err)

            exc.result = result
            exc.cnx_new = cnx_new
            raise exc

        return cnx_new

    def get_sane_defaults(self, adapter_name):
        """Pass."""
        return CNX_SANE_DEFAULTS.get(adapter_name, CNX_SANE_DEFAULTS["all"])

    def get_by_adapter(self, adapter_name, adapter_node=DEFAULT_NODE):
        """Get connections from an adapter."""
        adapter = self.parent.get_by_name(name=adapter_name, node=adapter_node)
        cnxs = adapter["cnx"]

        for cnx in cnxs:
            cnx["schemas"] = adapter["schemas"]["cnx"]

        return cnxs

    def get_by_key(
        self,
        value,
        value_key,
        adapter_name,
        adapter_node=DEFAULT_NODE,
        retry=0,
        sleep=1,
    ):
        """Pass."""
        tries = 1
        cnxs = self.get_by_adapter(adapter_name=adapter_name, adapter_node=adapter_node)
        while True:
            for cnx in cnxs:
                if cnx[value_key] == value:
                    return cnx

            tries += 1

            if tries > retry:
                break

            time.sleep(sleep)

            cnxs = self.get_by_adapter(
                adapter_name=adapter_name, adapter_node=adapter_node
            )

        value_key = value_key.upper()
        err = (
            f"No connection found on adapter {adapter_name!r} node {adapter_node!r} "
            f"with {value_key} of {value!r}"
        )
        raise NotFoundError(tablize_cnxs(cnxs=cnxs, err=err))

    def get_by_uuid(self, cnx_uuid, adapter_name, adapter_node=DEFAULT_NODE, **kwargs):
        """Pass."""
        kwargs["value_key"] = "uuid"
        return self.get_by_key(
            value=cnx_uuid,
            adapter_name=adapter_name,
            adapter_node=adapter_node,
            **kwargs,
        )

    def get_by_id(self, cnx_id, adapter_name, adapter_node=DEFAULT_NODE, **kwargs):
        """Pass."""
        kwargs["value_key"] = "id"
        return self.get_by_key(
            value=cnx_id, adapter_name=adapter_name, adapter_node=adapter_node, **kwargs
        )

    def test_by_id(self, cnx_id, adapter_name, adapter_node=DEFAULT_NODE):
        """Pass."""
        cnx_test = self.get_by_id(
            cnx_id=cnx_id, adapter_name=adapter_name, adapter_node=adapter_node
        )
        return self.test_cnx(cnx_test=cnx_test)

    def test_cnx(self, cnx_test, **kwargs):
        """Pass."""
        test_adapter_meta = {
            "adapter_name": cnx_test["adapter_name"],
            "adapter_name_raw": cnx_test["adapter_name_raw"],
            "adapter_node_id": cnx_test["node_id"],
            "adapter_node_name": cnx_test["node_name"],
            "cnx_schemas": cnx_test["schemas"],
        }

        test_old_config = cnx_test["config"]

        return self.do_test(
            test_adapter_meta=test_adapter_meta,
            test_old_config=test_old_config,
            **kwargs,
        )

    def test(self, adapter_name, adapter_node=DEFAULT_NODE, **kwargs):
        """Pass."""
        adapter = self.parent.get_by_name(name=adapter_name, node=adapter_node)

        test_adapter_meta = {
            "adapter_name": adapter["name"],
            "adapter_name_raw": adapter["name_raw"],
            "adapter_node_id": adapter["node_id"],
            "adapter_node_name": adapter["node_name"],
            "cnx_schemas": adapter["schemas"]["cnx"],
        }

        test_old_config = {}

        return self.do_test(
            test_adapter_meta=test_adapter_meta,
            test_old_config=test_old_config,
            **kwargs,
        )

    def do_test(self, test_adapter_meta, test_old_config, **kwargs):
        """Pass."""
        kwargs_config = kwargs.pop("kwargs_config", {})
        kwargs.update(kwargs_config)
        adapter_name = test_adapter_meta["adapter_name"]
        adapter_name_raw = test_adapter_meta["adapter_name_raw"]
        adapter_node_id = test_adapter_meta["adapter_node_id"]
        adapter_node_name = test_adapter_meta["adapter_node_name"]
        cnx_schemas = test_adapter_meta["cnx_schemas"]

        source = f"reachability test for adapter {adapter_name!r}"

        new_config = self.build_config(
            cnx_schemas=cnx_schemas,
            old_config=test_old_config,
            new_config=kwargs,
            source=source,
            adapter_name=adapter_name,
            adapter_node=adapter_node_name,
        )

        config_empty(schemas=cnx_schemas, new_config=new_config, source=source)

        result = self._test(
            adapter_name_raw=adapter_name_raw,
            adapter_node_id=adapter_node_id,
            config=new_config,
        )

        rtext = (result.text or "").strip()
        rjson = json_load(obj=rtext, error=False) if rtext else {}

        had_error = not result.ok or bool(rtext)

        if had_error:
            if rjson.get("type") == "AttributeError":
                err = (
                    f"Reachability test failed due to settings required for "
                    "testing reachability (supply at least hostname/domain/etc)"
                )
                msg = tablize_schemas(schemas=cnx_schemas, err=err)
                raise ConfigRequired(msg)
            else:
                rkw = ["{}: {}".format(k, v) for k, v in rjson.items()]
                rkw = "\n  " + "\n  ".join(rkw)
                msg = f"Reachability test failed:{rkw}"
                raise CnxTestError(msg)
        return rtext

    def update_cnx(self, cnx_update, **kwargs):
        """Pass."""
        kwargs_config = kwargs.pop("kwargs_config", {})
        kwargs.update(kwargs_config)

        adapter_name = cnx_update["adapter_name"]
        adapter_name_raw = cnx_update["adapter_name_raw"]
        adapter_node_name = cnx_update["node_name"]
        adapter_node_id = cnx_update["node_id"]
        cnx_schemas = cnx_update["schemas"]

        old_config = cnx_update["config"]
        old_uuid = cnx_update["uuid"]
        old_id = cnx_update["id"]

        source = (
            f"updating settings for connection ID {old_id!r} on adapter {adapter_name!r}"
        )

        new_config = self.build_config(
            cnx_schemas=cnx_schemas,
            old_config=old_config,
            new_config=kwargs,
            source=source,
            adapter_name=adapter_name,
            adapter_node=adapter_node_name,
        )

        config_empty(schemas=cnx_schemas, new_config=new_config, source=source)

        config_unchanged(
            schemas=cnx_schemas,
            old_config=old_config,
            new_config=new_config,
            source=source,
        )

        result = self._update(
            adapter_name_raw=adapter_name_raw,
            adapter_node_id=adapter_node_id,
            new_config=new_config,
            cnx_uuid=old_uuid,
        )

        self.check_if_gone(
            result=result,
            cnx_id=old_id,
            adapter_name=adapter_name,
            adapter_node=adapter_node_name,
        )

        cnx_new = self.get_by_uuid(
            cnx_uuid=result["id"],
            adapter_name=adapter_name,
            adapter_node=adapter_node_name,
            retry=CNX_RETRY,
        )

        error_in_status = result.get("status", "") == "error"
        error_empty = bool(result.get("error", ""))

        if any([error_in_status, error_empty]):
            rkw = ["{}: {}".format(k, v) for k, v in result.items()]
            rkw = "\n  " + "\n  ".join(rkw)

            err = f"Connection was updated but had a failure connecting:{rkw}"
            exc = CnxUpdateError(err)

            exc.result = result
            exc.cnx_old = cnx_update
            exc.cnx_new = cnx_new

            raise exc

        return cnx_new

    def check_if_gone(self, result, cnx_id, adapter_name, adapter_node):
        """Pass."""
        message = result.get("message", "")
        if message == CNX_GONE:
            cnxs = self.get_by_adapter(
                adapter_name=adapter_name, adapter_node=adapter_node
            )
            err = f"Connection with ID {cnx_id!r} no longer exists!"
            raise CnxGoneError(tablize_cnxs(cnxs=cnxs, err=err))

    def update_by_id(self, cnx_id, adapter_name, adapter_node=DEFAULT_NODE, **kwargs):
        """Pass."""
        cnx_update = self.get_by_id(
            cnx_id=cnx_id, adapter_name=adapter_name, adapter_node=adapter_node
        )
        return self.update_cnx(cnx_update=cnx_update, **kwargs)

    def delete_cnx(self, cnx_delete, delete_entities=False):
        """Pass."""
        adapter_name_raw = cnx_delete["adapter_name_raw"]
        adapter_node_id = cnx_delete["node_id"]
        cnx_uuid = cnx_delete["uuid"]

        return self._delete(
            adapter_name_raw=adapter_name_raw,
            adapter_node_id=adapter_node_id,
            cnx_uuid=cnx_uuid,
            delete_entities=delete_entities,
        )

    def delete_by_id(
        self, cnx_id, adapter_name, adapter_node=DEFAULT_NODE, delete_entities=False
    ):
        """Pass."""
        cnx_delete = self.get_by_id(
            cnx_id=cnx_id, adapter_name=adapter_name, adapter_node=adapter_node
        )
        return self.delete_cnx(cnx_delete=cnx_delete, delete_entities=delete_entities)

    def build_config(
        self, cnx_schemas, old_config, new_config, source, adapter_name, adapter_node,
    ):
        """Pass."""
        callbacks = {
            "cb_file": self.cb_file_upload,
            "adapter_name": adapter_name,
            "adapter_node": adapter_node,
        }

        config_unknown(
            schemas=cnx_schemas,
            new_config=new_config,
            source=source,
            callbacks=callbacks,
        )

        new_config = config_build(
            schemas=cnx_schemas,
            old_config=old_config,
            new_config=new_config,
            source=source,
            callbacks=callbacks,
        )

        return new_config

    def cb_file_upload(self, value, schema, callbacks, source):
        """Pass."""
        adapter_name = callbacks["adapter_name"]
        adapter_node = callbacks["adapter_node"]
        field_name = schema["name"]

        if isinstance(value, str):
            check = pathlib.Path(value).expanduser().resolve()
            if check.is_file():
                value = check

        if isinstance(value, pathlib.Path):
            value = value.expanduser().resolve()
            if not value.is_file():
                sinfo = config_info(schema=schema, value=str(value), source=source)
                raise ConfigInvalidValue(f"{sinfo}\nFile does not exist!")

            file_name = value.name
            file_content = value.read_text()
        elif hasattr(value, "read"):
            file_content = value.read()
            file_name = file_content[:20]
        else:
            sinfo = config_info(schema=schema, value=str(value), source=source)
            raise ConfigInvalidValue(
                f"{sinfo}\nFile is not an existing file or a file-like object!"
            )

        return self.parent.file_upload(
            name=adapter_name,
            field_name=field_name,
            file_name=file_name,
            file_content=file_content,
            node=adapter_node,
        )

    def _add(self, adapter_name_raw, adapter_node_id, new_config):
        """Direct API method to add a connection to an adapter.

        Args:
            adapter (:obj:`str`): name of adapter
            node_id (:obj:`str`): id of node running **adapter**
            config (:obj:`dict`): configuration values for new connection

        Returns:
            :obj:`str`: an empty str
        """
        data = {}
        data.update(new_config)
        data["instanceName"] = adapter_node_id

        path = self.parent.router.cnxs.format(adapter_name_raw=adapter_name_raw)

        return self.parent.request(
            method="put",
            path=path,
            json=data,
            error_json_bad_status=False,
            error_status=False,
        )

    def _test(self, adapter_name_raw, adapter_node_id, config):
        """Direct API method to add a connection to an adapter.

        Args:
            adapter (:obj:`str`): name of adapter
            node_id (:obj:`str`): id of node running **adapter**
            config (:obj:`dict`): configuration values to test reachability for

        Returns:
            :obj:`str`: an empty str
        """
        data = {}
        data.update(config)
        data["instanceName"] = adapter_node_id
        data["oldInstanceName"] = adapter_node_id

        path = self.parent.router.cnxs.format(adapter_name_raw=adapter_name_raw)

        return self.parent.request(method="post", path=path, json=data, raw=True)

    def _delete(
        self, adapter_name_raw, adapter_node_id, cnx_uuid, delete_entities=False
    ):
        """Direct API method to delete a connection from an adapter.

        Args:
            name_raw (:obj:`str`): name_raw of adapter
            node_id (:obj:`str`): id of node running **adapter**
            uuid (:obj:`str`): uuid of connection to delete
            delete_entities (:obj:`bool`, optional): default ``False`` -

                * if ``True`` delete the connection and also delete all asset entities
                  fetched by this connection
                * if ``False`` just delete the connection

        Returns:
            :obj:`str`: an empty str
        """
        data = {}
        data["instanceName"] = adapter_node_id

        params = {"deleteEntities": delete_entities}

        path = self.parent.router.cnxs_uuid.format(
            adapter_name_raw=adapter_name_raw, cnx_uuid=cnx_uuid
        )

        return self.parent.request(
            method="delete",
            path=path,
            json=data,
            params=params,
            error_json_bad_status=False,
            error_status=False,
        )

    def _update(self, adapter_name_raw, adapter_node_id, new_config, cnx_uuid):
        """Direct API method to update a connection on an adapter.

        Args:
            name_raw (:obj:`str`): name_raw of adapter
            node_id (:obj:`str`): id of node running **adapter**
            config (:obj:`dict`): configuration of connection to update
            uuid (:obj:`str`): uuid of connection to update

        Returns:
            :obj:`str`: an empty str
        """
        data = {}
        data.update(new_config)
        data["instanceName"] = adapter_node_id
        data["oldInstanceName"] = adapter_node_id

        path = self.parent.router.cnxs_uuid.format(
            adapter_name_raw=adapter_name_raw, cnx_uuid=cnx_uuid
        )
        return self.parent.request(
            method="put",
            path=path,
            json=data,
            error_json_bad_status=False,
            error_status=False,
        )



# -*- coding: utf-8 -*-
"""Test suite."""

import pytest
from axonius_api_client.exceptions import ApiError, NotFoundError

# from ...meta import TEST_PERM, TEST_ROLE

GUI_SECTION_WITH_SUBS = "system_settings"
GUI_SECTION_NO_SUBS = "ldap_login_settings"
GUI_NON_SUB_SECTION = "multiLine"
GUI_SUB_SECTION = "percentageThresholds"
GUI_SUB_KEYS = ["error", "success", "warning"]


class SettingsBasePublic:
    """Pass."""

    def val_schemas(self, schemas, config):
        """Pass."""
        for name, schema in schemas.items():
            assert schema["name"] == name
            assert schema["name"] in config
            assert isinstance(schema["required"], bool)

    def val_sub_section(self, name, meta, settings):
        """Pass."""
        assert isinstance(meta, dict) and meta
        assert meta["settings_title"] == settings["settings_title"]
        assert meta["name"] == name
        assert isinstance(meta["title"], str) and meta["title"]
        assert isinstance(meta["schemas"], dict) and meta["schemas"]
        assert isinstance(meta["sub_sections"], dict) and not meta["sub_sections"]
        assert isinstance(meta["parent_name"], str) and meta["parent_name"]
        assert isinstance(meta["parent_title"], str) and meta["parent_title"]

        assert isinstance(meta["config"], dict)
        for k, v in meta["config"].items():
            assert not isinstance(v, dict)

        self.val_schemas(schemas=meta["schemas"], config=meta["config"])

    def val_section(self, name, meta, settings):
        """Pass."""
        assert isinstance(meta, dict) and meta
        assert meta["settings_title"] == settings["settings_title"]
        assert meta["name"] == name
        assert isinstance(meta["title"], str) and meta["title"]
        assert isinstance(meta["schemas"], dict) and meta["schemas"]
        assert isinstance(meta["sub_sections"], dict)
        assert isinstance(meta["parent_name"], str) and not meta["parent_name"]
        assert isinstance(meta["parent_title"], str) and not meta["parent_title"]

        assert isinstance(meta["config"], dict)

        self.val_schemas(schemas=meta["schemas"], config=meta["config"])
        for sub_name, sub_meta in meta["sub_sections"].items():
            self.val_sub_section(name=sub_name, meta=sub_meta, settings=settings)

    def test_get(self, apiobj):
        """Pass."""
        settings = apiobj.get()
        assert isinstance(settings, dict)

        assert isinstance(settings["settings_title"], str) and settings["settings_title"]
        assert isinstance(settings["sections"], dict) and settings["sections"]
        assert isinstance(settings["config"], dict) and settings["config"]

        for name, meta in settings["sections"].items():
            self.val_section(name=name, meta=meta, settings=settings)
            assert name in settings["config"]


class TestSettingsGui(SettingsBasePublic):
    """Pass."""

    @pytest.fixture(scope="class")
    def apiobj(self, api_system):
        """Pass."""
        return api_system.settings_gui

    def test_get_section_full_config_true(self, apiobj):
        """Pass."""
        result = apiobj.get_section(section=GUI_SECTION_WITH_SUBS, full_config=True)
        assert isinstance(result, dict)
        assert "full_config" in result

    def test_get_sub_section_full_config_true(self, apiobj):
        """Pass."""
        result = apiobj.get_sub_section(
            section=GUI_SECTION_WITH_SUBS, sub_section=GUI_SUB_SECTION, full_config=True
        )
        assert isinstance(result, dict)
        assert "full_config" in result

    def test_get_section_full_config_false(self, apiobj):
        """Pass."""
        result = apiobj.get_section(section=GUI_SECTION_WITH_SUBS, full_config=False)
        assert isinstance(result, dict)
        assert "full_config" not in result

    def test_get_sub_section_full_config_false(self, apiobj):
        """Pass."""
        result = apiobj.get_sub_section(
            section=GUI_SECTION_WITH_SUBS, sub_section=GUI_SUB_SECTION, full_config=False
        )
        assert isinstance(result, dict)
        assert "full_config" not in result

    def test_get_section_error(self, apiobj):
        """Pass."""
        with pytest.raises(NotFoundError):
            apiobj.get_section(section="badwolf")

    def test_get_sub_section_error(self, apiobj):
        """Pass."""
        with pytest.raises(NotFoundError):
            apiobj.get_sub_section(section=GUI_SECTION_WITH_SUBS, sub_section="badwolf")

    def test_get_sub_section_no_subsections(self, apiobj):
        """Pass."""
        with pytest.raises(ApiError):
            apiobj.get_sub_section(section=GUI_SECTION_NO_SUBS, sub_section="badwolf")

    def test_update_section(self, apiobj):
        """Pass."""
        old_section = apiobj.get_section(section=GUI_SECTION_WITH_SUBS)
        old_config = old_section["config"]
        old_value = old_config[GUI_NON_SUB_SECTION]

        update_value = not old_value

        new_section_args = {GUI_NON_SUB_SECTION: update_value}
        new_section = apiobj.update_section(
            section=GUI_SECTION_WITH_SUBS, **new_section_args
        )
        new_value = new_section["config"][GUI_NON_SUB_SECTION]
        assert new_value == update_value and old_value != new_value

        reset_section_args = {GUI_NON_SUB_SECTION: old_value}
        reset_section = apiobj.update_section(
            section=GUI_SECTION_WITH_SUBS, **reset_section_args
        )
        reset_value = reset_section["config"][GUI_NON_SUB_SECTION]
        assert reset_value == old_value and reset_value != new_value

    def test_update_sub_section(self, apiobj):
        """Pass."""
        old_section = apiobj.get_sub_section(
            section=GUI_SECTION_WITH_SUBS, sub_section=GUI_SUB_SECTION
        )
        sub_key = GUI_SUB_KEYS[0]
        old_config = old_section["config"]
        old_value = old_config[sub_key]
        update_value = old_value + 1

        new_section_args = {sub_key: update_value}
        new_section = apiobj.update_sub_section(
            section=GUI_SECTION_WITH_SUBS,
            sub_section=GUI_SUB_SECTION,
            **new_section_args
        )
        new_value = new_section["config"][sub_key]
        assert new_value == update_value and old_value != new_value

        reset_section_args = {sub_key: old_value}
        reset_section = apiobj.update_sub_section(
            section=GUI_SECTION_WITH_SUBS,
            sub_section=GUI_SUB_SECTION,
            **reset_section_args
        )
        reset_value = reset_section["config"][sub_key]
        assert reset_value == old_value and reset_value != new_value


class TestSettingsCore(SettingsBasePublic):
    """Pass."""

    @pytest.fixture(scope="class")
    def apiobj(self, api_system):
        """Pass."""
        return api_system.settings_core


class TestSettingsLifecycle(SettingsBasePublic):
    """Pass."""

    @pytest.fixture(scope="class")
    def apiobj(self, api_system):
        """Pass."""
        return api_system.settings_lifecycle



# -*- coding: utf-8 -*-
"""API models for working with device and user assets."""
from ..routers import API_VERSION
from .asset_mixin import AssetMixin


class Users(AssetMixin):
    """User related API methods."""

    NO_CURSOR_OVERRIDE = True
    # cursor in 3.2 only fetches device assets, will be fixed in 3.3
    FIELD_USERNAME = "specific_data.data.username"
    FIELD_MAIL = "specific_data.data.mail"

    @property
    def fields_default(self):
        """Fields to use by default for getting assets."""
        return self.FIELDS_API + [self.FIELD_USERNAME, self.FIELD_MAIL]

    @property
    def router(self):
        """Router for this API model."""
        return API_VERSION.users

    def get_by_usernames(self, values, **kwargs):
        """Build a query to get assets where username in values."""
        kwargs["field"] = self.FIELD_USERNAME
        kwargs["field_manual"] = True
        kwargs["values"] = values
        return self.get_by_values(**kwargs)

    def get_by_username_regex(self, value, **kwargs):
        """Build a query to get assets where username regex matches value."""
        kwargs["field"] = self.FIELD_USERNAME
        kwargs["field_manual"] = True
        kwargs["value"] = value
        return self.get_by_value_regex(**kwargs)

    def get_by_username(self, value, **kwargs):
        """Build a query to get assets where username == value."""
        kwargs["field"] = self.FIELD_USERNAME
        kwargs["field_manual"] = True
        kwargs["value"] = value
        return self.get_by_value(**kwargs)

    def get_by_mails(self, values, **kwargs):
        """Build a query to get assets where mail in values."""
        kwargs["field"] = self.FIELD_MAIL
        kwargs["field_manual"] = True
        kwargs["values"] = values
        return self.get_by_values(**kwargs)

    def get_by_mail_regex(self, value, **kwargs):
        """Build a query to get assets where mail regex matches value."""
        kwargs["field"] = self.FIELD_MAIL
        kwargs["field_manual"] = True
        kwargs["value"] = value
        return self.get_by_value_regex(**kwargs)

    def get_by_mail(self, value, **kwargs):
        """Build a query to get assets where mail == value."""
        kwargs["field"] = self.FIELD_MAIL
        kwargs["field_manual"] = True
        kwargs["value"] = value
        return self.get_by_value(**kwargs)



# -*- coding: utf-8 -*-
"""Command line interface for Axonius API Client."""
from ...context import CONTEXT_SETTINGS, click
from ...options import AUTH, NODE_CNX, SPLIT_CONFIG_OPT, add_options
from .grp_common import EXPORT, PROMPTS, add_cnx, prompt_config

OPTIONS = [
    *AUTH,
    EXPORT,
    *NODE_CNX,
    *PROMPTS,
    SPLIT_CONFIG_OPT,
]


@click.command(name="add", context_settings=CONTEXT_SETTINGS)
@add_options(OPTIONS)
@click.pass_context
def cmd(ctx, url, key, secret, config, **kwargs):
    """Add a connection from prompts or arguments."""
    client = ctx.obj.start_client(url=url, key=key, secret=secret)
    new_config = dict(config)
    prompt_config(ctx=ctx, client=client, new_config=new_config, **kwargs)
    add_cnx(ctx=ctx, client=client, new_config=new_config, **kwargs)



# -*- coding: utf-8 -*-
"""Command line interface for Axonius API Client."""
from ...context import CONTEXT_SETTINGS, click
from ...options import AUTH, add_options
from .grp_common import EXPORT_FORMAT, handle_export

OPTIONS = [
    *AUTH,
    EXPORT_FORMAT,
    click.option(
        "--tag",
        "-t",
        "value",
        help="Tags of saved queries",
        required=True,
        multiple=True,
        show_envvar=True,
        show_default=True,
    ),
]


@click.command(name="get-by-tags", context_settings=CONTEXT_SETTINGS)
@add_options(OPTIONS)
@click.pass_context
def cmd(ctx, url, key, secret, export_format, **kwargs):
    """Get saved queries."""
    client = ctx.obj.start_client(url=url, key=key, secret=secret)

    p_grp = ctx.parent.parent.command.name
    apiobj = getattr(client, p_grp)

    with ctx.obj.exc_wrap(wraperror=ctx.obj.wraperror):
        rows = apiobj.saved_query.get_by_tags(**kwargs)

    handle_export(ctx=ctx, rows=rows, export_format=export_format)



# -*- coding: utf-8 -*-
"""Version information for this package."""
__version__ = "3.0.4"
VERSION = __version__
""":obj:`str`: Version of package."""

__url__ = "https://github.com/Axonius/axonius_api_client"
URL = __url__
""":obj:`str`: URL of package."""

__author__ = "Axonius"
AUTHOR = __author__
""":obj:`str`: Auth of package."""

__title__ = "axonius_api_client"
TITLE = __title__
""":obj:`str`: Title of package."""

__project__ = "axonius_api_client"
PROJECT = __project__
""":obj:`str`: Name of package."""

__author_email__ = "support@axonius.com"
AUTHOR_EMAIL = __author_email__
""":obj:`str`: Author email of package."""

__description__ = "Axonius API client for Python"
DESCRIPTION = __description__
""":obj:`str`: Description of package."""

__docs__ = "https://axonius-api-client.readthedocs.io/en/latest/"
DOCS = __docs__
""":obj:`str`: Link to the documentation for this package."""

__license__ = "MIT"
LICENSE = __license__
""":obj:`str`: License of package."""

__copyright__ = "Copyright Axonius 2020"
COPYRIGHT = __copyright__
""":obj:`str`: Copyright of package."""



# -*- coding: utf-8 -*-
"""Test suite."""
from axonius_api_client.api.parsers.tables import tab_map, tablize
from axonius_api_client.constants import KEY_MAP_SCHEMA


def test_tablize_footer_true():
    """Pass."""
    key = "badwolf1"
    key_value = "badwolf2"
    err = "badwolf3"

    exp = [err, "", key, "----------", key_value, "----------", key, "", err]

    value = [{key: key_value}]
    result = tablize(value=value, err=err, fmt="simple", footer=True)
    lines = result.splitlines()
    assert lines == exp


def test_tablize_footer_false():
    """Pass."""
    key = "badwolf1"
    key_value = "badwolf2"
    err = "badwolf3"

    exp = [err, "", key, "----------", key_value, "", err]

    value = [{key: key_value}]
    result = tablize(value=value, err=err, fmt="simple", footer=False)
    lines = result.splitlines()
    assert lines == exp


def test_tablize_err_false():
    """Pass."""
    key = "badwolf1"
    key_value = "badwolf2"

    exp = ["", "", key, "----------", key_value, "----------", key, ""]

    value = [{key: key_value}]
    result = tablize(value=value, err=None, fmt="simple", footer=True)
    lines = result.splitlines()
    assert lines == exp


def test_tab_map_orig_false():
    """Pass."""
    value = {"name": "schema_str", "type": "string", "required": False, "other": "bloop"}
    exp = {"Name": "schema_str", "Type": "string", "Required": False}
    result = tab_map(value=value, key_map=KEY_MAP_SCHEMA, orig=False)
    assert result == exp


def test_tab_map_orig_true():
    """Pass."""
    value = {"name": "schema_str", "type": "string", "required": False, "other": "bloop"}
    exp = {"Name": "schema_str", "Type": "string", "Required": False, "other": "bloop"}
    result = tab_map(value=value, key_map=KEY_MAP_SCHEMA, orig=True)
    assert result == exp


def test_tab_map_list():
    """Pass."""
    long_str = ("badwolf " * 30).split(" ")
    long_str_exp = "\n".join(long_str)
    value = {
        "name": long_str,
        "type": "string",
        "required": False,
    }
    exp = {
        "Name": long_str_exp,
        "Required": False,
        "Type": "string",
    }
    result = tab_map(value=value, key_map=KEY_MAP_SCHEMA)
    assert result == exp


def test_tab_map_long_str():
    """Pass."""
    long_str = "badwolf " * 30
    long_str_exp = (
        "badwolf badwolf badwolf\n"
        "badwolf badwolf badwolf\n"
        "badwolf badwolf badwolf\n"
        "badwolf badwolf badwolf\n"
        "badwolf badwolf badwolf\n"
        "badwolf badwolf badwolf\n"
        "badwolf badwolf badwolf\n"
        "badwolf badwolf badwolf\n"
        "badwolf badwolf badwolf\n"
        "badwolf badwolf badwolf"
    )
    value = {
        "title": long_str,
        "type": "string",
        "required": False,
    }
    exp = {
        "Title": long_str_exp,
        "Required": False,
        "Type": "string",
    }
    result = tab_map(value=value, key_map=KEY_MAP_SCHEMA)
    assert result == exp


def test_tab_map_list_other():
    """Pass."""
    value = {
        "name": "schema_str",
        "type": "string",
        "required": False,
        "other": ["bloop", "floop"],
    }
    exp = {
        "Name": "schema_str",
        "Type": "string",
        "Required": False,
        "other": "bloop\nfloop",
    }
    result = tab_map(value=value, key_map=KEY_MAP_SCHEMA, orig=True)
    assert result == exp



# -*- coding: utf-8 -*-
"""Test suite for axonius_api_client.auth."""
import axonius_api_client as axonapi


def test_api_v1_routers():
    """Test api.routers.ApiV1."""
    api_routes = axonapi.api.routers.ApiV1
    for obj in api_routes.all_objects:
        assert isinstance(obj, axonapi.api.routers.Router)
        assert obj._object_type in format(obj)
        assert obj._object_type in repr(obj)
        for route in obj._routes:
            assert hasattr(obj, route)



# -*- coding: utf-8 -*-
"""Command line interface for Axonius API Client."""
from ...tools import json_dump
from ..context import CONTEXT_SETTINGS, click
from ..options import AUTH, add_options

DESTROY = click.option(
    "--destroy/--no-destroy",
    "-d/-nd",
    "destroy",
    help="Actually perform the destroy operation",
    default=None,
    is_flag=True,
    show_envvar=True,
    show_default=True,
    required=True,
)
HISTORY = click.option(
    "--history/--no-history",
    "-h/-nh",
    "history",
    help="Also delete all historical asset records",
    default=None,
    is_flag=True,
    show_envvar=True,
    show_default=True,
    required=True,
)
OPTIONS = [
    *AUTH,
    DESTROY,
    HISTORY,
]


@click.command(name="destroy", context_settings=CONTEXT_SETTINGS)
@add_options(OPTIONS)
@click.pass_context
def cmd(ctx, url, key, secret, **kwargs):
    """Delete ALL asset records of this type from the database."""
    client = ctx.obj.start_client(url=url, key=key, secret=secret)

    p_grp = ctx.parent.command.name
    apiobj = getattr(client, p_grp)

    with ctx.obj.exc_wrap(wraperror=ctx.obj.wraperror):
        data = apiobj.destroy(**kwargs)

    click.secho(json_dump(data))



# -*- coding: utf-8 -*-
"""Test suite for axonapi.api.assets."""
import copy
import json

import pytest
from axonius_api_client.api.assets.saved_query import check_gui_page_size
from axonius_api_client.constants import GUI_PAGE_SIZES, SIMPLE
from axonius_api_client.exceptions import ApiError, NotFoundError

from ...meta import QUERIES


def load_test_data(apiobj):
    """Pass."""
    apiobj.TEST_DATA = getattr(apiobj, "TEST_DATA", {})

    if not apiobj.TEST_DATA.get("saved_queries"):
        apiobj.TEST_DATA["saved_queries"] = apiobj.saved_query.get()

    return apiobj


def test_check_gui_page_size_error():
    """Pass."""
    gui_page_size = 9999
    with pytest.raises(ApiError):
        check_gui_page_size(size=gui_page_size)


class SavedQueryPrivate:
    """Pass."""

    def test_private_get(self, apiobj):
        """Pass."""
        result = apiobj.saved_query._get()
        assert isinstance(result, dict)

        rows = result["assets"]
        assert isinstance(rows, list)

        for row in rows:
            assert isinstance(row, dict)
            validate_sq(row)


class SavedQueryPublic:
    """Pass."""

    def test_get_no_generator(self, apiobj):
        """Pass."""
        rows = apiobj.saved_query.get(generator=False)
        assert not rows.__class__.__name__ == "generator"
        assert isinstance(rows, list)
        for row in rows:
            assert isinstance(row, dict)
            validate_sq(row)

    def test_get_generator(self, apiobj):
        """Pass."""
        gen = apiobj.saved_query.get(generator=True)
        assert gen.__class__.__name__ == "generator"
        assert not isinstance(gen, list)

        rows = [x for x in gen]
        for row in rows:
            assert isinstance(row, dict)
            validate_sq(row)

    def test_get_max_pages(self, apiobj):
        """Pass."""
        rows = apiobj.saved_query.get(max_pages=1, page_size=2)
        assert isinstance(rows, list)
        assert len(rows) == 2
        for row in rows:
            assert isinstance(row, dict)
            validate_sq(row)

    def test_get_max_rows(self, apiobj):
        """Pass."""
        rows = apiobj.saved_query.get(max_rows=1)
        assert isinstance(rows, list)
        assert len(rows) == 1
        for row in rows:
            assert isinstance(row, dict)
            validate_sq(row)

    def test_get_tags(self, apiobj):
        """Pass."""
        tags = apiobj.saved_query.get_tags()
        assert isinstance(tags, list)
        for tag in tags:
            assert isinstance(tag, str)

    def test_get_by_name(self, apiobj):
        """Pass."""
        sq = apiobj.TEST_DATA["saved_queries"][0]
        value = sq["name"]
        row = apiobj.saved_query.get_by_name(value=value)
        assert isinstance(row, dict)
        assert row["name"] == value

    def test_get_by_name_error(self, apiobj):
        """Pass."""
        value = "badwolf_zzzzz"
        with pytest.raises(NotFoundError):
            apiobj.saved_query.get_by_name(value=value)

    def test_get_by_uuid(self, apiobj):
        """Pass."""
        sq = apiobj.TEST_DATA["saved_queries"][0]
        value = sq["uuid"]
        row = apiobj.saved_query.get_by_uuid(value=value)
        assert isinstance(row, dict)
        assert row["uuid"] == value

    def test_get_by_uuid_error(self, apiobj):
        """Pass."""
        value = "badwolf_zzzzz"
        with pytest.raises(NotFoundError):
            apiobj.saved_query.get_by_uuid(value=value)

    def test_get_by_tags(self, apiobj):
        """Pass."""
        tags = [y for x in apiobj.TEST_DATA["saved_queries"] for y in x.get("tags", [])]
        value = tags[0]
        rows = apiobj.saved_query.get_by_tags(value=value)
        assert isinstance(rows, list)
        for row in rows:
            assert isinstance(row, dict)
            assert value in row["tags"]

    def test_get_by_tags_error(self, apiobj):
        """Pass."""
        value = "badwolf_zzzzz"
        with pytest.raises(NotFoundError):
            apiobj.saved_query.get_by_tags(value=value)

    @pytest.fixture(scope="class")
    def sq_fixture(self, apiobj):
        """Pass."""
        field_simple = apiobj.TEST_DATA["field_simple"]

        name = "badwolf torked"
        fields = ["adapters", "last_seen", "id", field_simple]
        exp_fields = [
            "adapters",
            "specific_data.data.last_seen",
            "specific_data.data.id",
            field_simple,
        ]
        sort_field = field_simple
        colfilters = {field_simple: "a"}
        sort_desc = False
        gui_page_size = GUI_PAGE_SIZES[-1]
        tags = ["badwolf1", "badwolf2"]
        description = "badwolf torked"
        query = QUERIES["not_last_seen_day"]

        try:
            apiobj.saved_query.delete_by_name(value=name)
        except NotFoundError:
            pass

        row_original = apiobj.saved_query.add(
            name=name,
            fields=fields,
            sort_field=sort_field,
            sort_descending=sort_desc,
            column_filters=colfilters,
            gui_page_size=gui_page_size,
            tags=tags,
            description=description,
            query=query,
        )
        assert isinstance(row_original, dict)
        row_validate = copy.deepcopy(row_original)
        row = copy.deepcopy(row_original)
        validate_sq(row_validate)

        row_name = row.pop("name")
        assert row_name == name

        row_query_type = row.pop("query_type")
        assert row_query_type == "saved"

        row_tags = row.pop("tags")
        assert row_tags == tags

        row_description = row.pop("description")
        assert row_description == description

        row_view = row.pop("view")
        assert isinstance(row_view, dict)

        row_archived = row.pop("archived")
        assert row_archived is False

        row_date_fetched = row.pop("date_fetched")
        assert isinstance(row_date_fetched, str)

        row_user_id = row.pop("user_id")
        assert isinstance(row_user_id, str)

        row_last_updated = row.pop("last_updated")
        assert isinstance(row_last_updated, str)

        row_uuid = row.pop("uuid")
        assert isinstance(row_uuid, str)

        row_updated_by = row.pop("updated_by")
        assert isinstance(row_updated_by, str)

        row_updated_by_json = json.loads(row_updated_by)
        assert isinstance(row_updated_by_json, dict)

        assert not row

        # row["view"]
        row_view_sort = row_view.pop("sort")
        assert isinstance(row_view_sort, dict)

        row_view_fields = row_view.pop("fields")
        assert isinstance(row_view_fields, list)
        assert row_view_fields == exp_fields

        row_view_query = row_view.pop("query")
        assert isinstance(row_view_query, dict)

        row_view_page_size = row_view.pop("pageSize")
        assert row_view_page_size == gui_page_size

        row_view_colfilters = row_view.pop("colFilters")
        assert row_view_colfilters == colfilters

        assert not row_view

        # row["view"]["sort"]
        row_view_sort_field = row_view_sort.pop("field")
        assert row_view_sort_field == sort_field

        row_view_sort_desc = row_view_sort.pop("desc")
        assert row_view_sort_desc == sort_desc

        assert not row_view_sort

        # row["view"]["query"]
        row_view_query_filter = row_view_query.pop("filter")
        assert row_view_query_filter == query

        row_view_query_expressions = row_view_query.pop("expressions")
        assert row_view_query_expressions == []

        # XXX TBD
        # row_view_query_search = row_view_query.pop("search")
        # assert row_view_query_search == ""

        assert not row_view_query
        yield row_original
        try:
            apiobj.saved_query.delete_by_name(value=name)
        except NotFoundError:
            pass

    def test_add_remove(self, apiobj, sq_fixture):
        """Pass."""
        row = apiobj.saved_query.delete_by_name(value=sq_fixture["name"])
        assert isinstance(row, dict)
        assert row["uuid"] == sq_fixture["uuid"]

        with pytest.raises(NotFoundError):
            apiobj.saved_query.get_by_name(value=sq_fixture["name"])

    def test_add_error_no_fields(self, apiobj):
        """Pass."""
        name = "badwolf_zzzzz"
        with pytest.raises(ApiError):
            apiobj.saved_query.add(name=name)

    def test_add_error_bad_sort_field(self, apiobj):
        """Pass."""
        name = "badwolf_zzzzz"
        fields = "last_seen"
        sort_field = "badwolf"
        with pytest.raises(ApiError):
            apiobj.saved_query.add(name=name, fields=fields, sort_field=sort_field)

    def test_add_error_bad_colfilter(self, apiobj):
        """Pass."""
        name = "badwolf_zzzzz"
        fields = "last_seen"
        colfilters = {"badwolf": "badwolf"}
        with pytest.raises(ApiError):
            apiobj.saved_query.add(name=name, fields=fields, column_filters=colfilters)

    # XXX not in rest api yet
    # def test_update_error(self, apiobj, sq_fixture):
    #     """Pass."""
    #     with pytest.raises(ApiError):
    #         apiobj.saved_query.update(row=sq_fixture)

    # def test_update_tags_add_remove(self, apiobj, sq_fixture):
    #     """Pass."""
    #     tags = ["badwolf_tango", "badwolf_foxtrot"]
    #     old_tags = sq_fixture.get("tags", [])
    #     exp = old_tags + tags
    #     add_result = apiobj.saved_query.update(row=sq_fixture, add_tags=tags)
    #     assert add_result["tags"] == exp


class TestSavedQueryDevices(SavedQueryPrivate, SavedQueryPublic):
    """Pass."""

    @pytest.fixture(scope="class")
    def apiobj(self, api_devices):
        """Pass."""
        return load_test_data(api_devices)


class TestSavedQueryUsers(SavedQueryPrivate, SavedQueryPublic):
    """Pass."""

    @pytest.fixture(scope="class")
    def apiobj(self, api_users):
        """Pass."""
        return load_test_data(api_users)


def validate_qexpr(qexpr, asset):
    """Pass."""
    assert isinstance(qexpr, dict)
    # XXX build a query builder :)

    compop = qexpr.pop("compOp")
    assert isinstance(compop, str)

    field = qexpr.pop("field")
    assert isinstance(field, str)

    idx = qexpr.pop("i", 0)
    assert isinstance(idx, int)

    leftbracket = qexpr.pop("leftBracket")
    assert isinstance(leftbracket, bool)

    rightbracket = qexpr.pop("rightBracket")
    assert isinstance(rightbracket, bool)

    logicop = qexpr.pop("logicOp")
    assert isinstance(logicop, str)

    notflag = qexpr.pop("not")
    assert isinstance(notflag, bool)

    value = qexpr.pop("value")
    assert isinstance(value, SIMPLE) or value is None

    obj = qexpr.pop("obj", False)
    assert isinstance(obj, bool)

    nesteds = qexpr.pop("nested", [])
    assert isinstance(nesteds, list)

    fieldtype = qexpr.pop("fieldType", "")
    assert isinstance(fieldtype, str)

    children = qexpr.pop("children", [])  # new in 2.15
    assert isinstance(children, list)  # new in 2.15

    filtered_adapters = qexpr.pop("filteredAdapters", {})
    assert isinstance(filtered_adapters, dict) or filtered_adapters is None

    context = qexpr.pop("context", "")  # new in 2.15
    assert isinstance(context, str)

    timestamp = qexpr.pop("timestamp", "")
    assert isinstance(timestamp, str)

    brackweight = qexpr.pop("bracketWeight", 0)
    assert isinstance(brackweight, int)

    qfilter = qexpr.pop("filter", "")
    assert isinstance(qfilter, str)

    for nested in nesteds:
        validate_nested(nested, asset)

    for child in children:
        validate_nested(child, asset)

    assert not qexpr, list(qexpr)


def validate_nested(nested, asset):
    """Pass."""
    assert isinstance(nested, dict)

    # new in 2.10, unsure of
    # if not None, dict with keys: clearAll, selectAll, selectedValues
    # XXX figure this out
    nfiltered_adapters = nested.pop("filteredAdapters", {})
    assert isinstance(nfiltered_adapters, dict) or nfiltered_adapters is None

    ncondition = nested.pop("condition")
    assert isinstance(ncondition, str)

    nexpr = nested.pop("expression")
    assert isinstance(nexpr, dict)

    nidx = nested.pop("i")
    assert isinstance(nidx, int)

    assert not nested, list(nested)


def validate_sq(asset):
    """Pass."""
    assert isinstance(asset, dict)

    original = copy.deepcopy(asset)
    assert original == asset

    assert asset["query_type"] in ["saved"]

    date_fetched = asset.pop("date_fetched")
    assert isinstance(date_fetched, str)

    last_updated = asset.pop("last_updated", "")
    assert isinstance(last_updated, str)

    name = asset.pop("name")
    assert isinstance(name, str)

    query_type = asset.pop("query_type")
    assert isinstance(query_type, str)

    user_id = asset.pop("user_id")
    assert isinstance(user_id, str)

    uuid = asset.pop("uuid")
    assert isinstance(uuid, str)

    description = asset.pop("description")
    assert isinstance(description, str) or description is None

    timestamp = asset.pop("timestamp", "")
    assert isinstance(timestamp, str)

    archived = asset.pop("archived", False)  # added in 2.15
    assert isinstance(archived, bool)

    updated_by_str = asset.pop("updated_by")
    assert isinstance(updated_by_str, str)

    updated_by = json.loads(updated_by_str)
    assert isinstance(updated_by, dict)

    updated_by_deleted = updated_by.pop("deleted")
    assert isinstance(updated_by_deleted, bool)

    updated_str_keys = ["username", "source", "first_name", "last_name"]
    for updated_str_key in updated_str_keys:
        val = updated_by.pop(updated_str_key)
        assert isinstance(val, str)

    assert not updated_by

    tags = asset.pop("tags", [])
    assert isinstance(tags, list)
    for tag in tags:
        assert isinstance(tag, str)

    predefined = asset.pop("predefined", False)
    assert isinstance(predefined, bool)

    view = asset.pop("view")
    assert isinstance(view, dict)

    colsizes = view.pop("coloumnSizes", [])
    assert isinstance(colsizes, list)
    # XXX figure this out

    for x in colsizes:
        assert isinstance(x, int)

    fields = view.pop("fields")
    assert isinstance(fields, list)

    for x in fields:
        assert isinstance(x, str)

    page = view.pop("page", 0)
    assert isinstance(page, int)

    pagesize = view.pop("pageSize", 0)
    assert isinstance(pagesize, int)

    sort = view.pop("sort")
    assert isinstance(sort, dict)

    sort_desc = sort.pop("desc")
    assert isinstance(sort_desc, bool)

    sort_field = sort.pop("field")
    assert isinstance(sort_field, str)

    query = view.pop("query")
    assert isinstance(query, dict)

    colfilters = view.pop("colFilters", {})
    assert isinstance(colfilters, dict)
    for k, v in colfilters.items():
        assert isinstance(k, str)
        assert isinstance(v, str)

    qfilter = query.pop("filter")
    assert isinstance(qfilter, str) or qfilter is None

    qexprs = query.pop("expressions", [])
    assert isinstance(qexprs, list)

    qmeta = query.pop("meta", {})
    assert isinstance(qmeta, dict)
    # XXX figure this out / parse more fully

    qonlyexprfilter = query.pop("onlyExpressionsFilter", "")
    assert isinstance(qonlyexprfilter, str)
    # XXX figure this out

    qsearch = query.pop("search", None)
    assert qsearch is None or isinstance(qsearch, str)
    # XXX figure this out

    historical = view.pop("historical", None)
    assert historical is None or isinstance(historical, SIMPLE)
    # XXX still used?

    for qexpr in qexprs:
        validate_qexpr(qexpr, asset)

    assert not query, list(query)
    assert not sort, list(sort)
    assert not view, list(view)
    assert not asset, list(asset)



# -*- coding: utf-8 -*-
"""API models for working with device and user assets."""
from ...constants import GUI_PAGE_SIZES
from ...exceptions import ApiError, NotFoundError
from ...tools import listify
from ..mixins import ChildMixins, PagingMixinsObject


def check_gui_page_size(size=None):
    """Pass."""
    if size:
        if size not in GUI_PAGE_SIZES:
            raise ApiError(
                f"gui_page_size of {size} is invalid, must be one of {GUI_PAGE_SIZES}"
            )
    else:
        size = GUI_PAGE_SIZES[0]
    return size


class SavedQuery(ChildMixins, PagingMixinsObject):
    """ChildMixins API model for working with saved queries for the parent asset type."""

    def get_by_tags(self, value, **kwargs):
        """Get saved queries by tags."""
        value = listify(value)
        rows = self.get(**kwargs)
        matches = []
        known = set()

        for row in rows:
            for tag in row.get("tags", []):
                known.add(tag)
                if tag in value and row not in matches:
                    matches.append(row)

        if not matches:
            valid = "\n  " + "\n  ".join(sorted(list(known)))
            msg = f"No saved query found with tags {value!r}, valid tags:{valid}"
            raise NotFoundError(msg)
        return matches

    def get_tags(self, **kwargs):
        """Get all tags for saved queries."""
        rows = self.get(**kwargs)
        tags = [y for x in rows for y in x.get("tags", [])]
        return sorted(list(set(tags)))

    def add(
        self,
        name,
        query=None,
        tags=None,
        description=None,
        fields=None,
        fields_manual=None,
        fields_default=False,
        sort_field=None,
        sort_descending=True,
        column_filters=None,
        gui_page_size=None,
        fields_map=None,
        **kwargs,
    ):
        """Create a saved query.

        Warning:
            Queries created with this method will NOT show the filters in the
            query wizard!

        Args:
            name (:obj:`str`): name of saved query to create
            query (:obj:`str`): query built by GUI query wizard
            fields (:obj:`object`): fields/columns
            fields_manual (:obj:`list` of :obj:`str`, optional): default ``None`` -
                list of fully qualified fields to include for each asset
            fields_regex (:obj:`list` of :obj:`str`, optional): default ``None`` -
                list of fields to add using regular expression matches, will be
                validated and process into the matching fully qualified names using
                :meth:`Fields.validate`
            fields_default (:obj:`bool`, optional): default ``True`` -
                Include the fields in _default_fields
            fields_error (:obj:`bool`, optional): default ``True`` -
                throw an exception if fields fail to be validated by
                :meth:`Fields.validate`
            sort (:obj:`str`, optional): default ``None`` - field to sort results on
            sort_descending (:obj:`bool`, optional): default ``True`` - sort on
                **field** in descending order
            column_filters (:obj:`dict`, optional): default ``None`` - column
                filters keyed as field_name:value
            gui_page_size (:obj:`int`, optional): default ``None`` -
                show N rows per page in GUI

        Returns:
            :obj:`dict`: metadata of saved query that was created
        """
        gui_page_size = check_gui_page_size(size=gui_page_size)
        fields_map = fields_map or self.parent.fields.get()

        fields = self.parent.fields.validate(
            fields=fields,
            fields_manual=fields_manual,
            fields_regex=None,
            fields_default=fields_default,
            fields_map=fields_map,
        )

        if sort_field:
            sort_field = self.parent.fields.get_field_name(
                value=sort_field, fields_map=fields_map
            )

        data_column_filters = {}
        if column_filters:
            for col_field, col_value in column_filters.items():
                col_field = self.parent.fields.get_field_name(
                    value=col_field, fields_map=fields_map
                )
                data_column_filters[col_field] = col_value

        data_query = {}
        data_query["filter"] = query
        data_query["expressions"] = []  # query wizard generated only
        # data_query["search"] = ""  # tbd

        data_sort = {}
        data_sort["desc"] = sort_descending
        data_sort["field"] = sort_field or ""

        data_view = {}
        data_view["query"] = data_query
        data_view["sort"] = data_sort
        data_view["fields"] = fields
        data_view["pageSize"] = gui_page_size
        data_view["colFilters"] = data_column_filters or {}

        data = {}
        data["name"] = name
        data["query_type"] = "saved"
        data["description"] = description
        data["view"] = data_view
        data["tags"] = tags or []

        added = self._add(data=data)
        kwargs["value"] = added
        return self.get_by_uuid(**kwargs)

    # XXX this isn't in API server yet
    '''
    def update(
        self,
        row,
        add_tags=None,
        remove_tags=None,
        add_fields=None,
        remove_fields=None,
        add_column_filters=None,
        remove_column_filters=None,
        description=None,
        sort_field=None,
        sort_descending=None,
        gui_page_size=None,
        **kwargs,
    ):
        """Update a saved query."""
        new_row = copy.deepcopy(row)
        old_tags = row.get("tags", [])
        old_fields = row["view"]["fields"]
        old_column_filters = row["view"].get("colFilters", {})

        if any([add_fields, remove_fields, sort_field]):
            fields_map = kwargs.get("fields_map") or self.parent.fields.get()

        if add_fields:
            add_fields = self.parent.fields.validate(
                fields=add_fields,
                fields_manual=None,
                fields_regex=None,
                fields_default=False,
                fields_map=fields_map,
            )
            new_fields = old_fields + [x for x in add_fields if x not in old_fields]
            row["view"]["fields"] = new_fields

        if remove_fields:
            remove_fields = self.parent.fields.validate(
                fields=remove_fields,
                fields_manual=None,
                fields_regex=None,
                fields_default=False,
                fields_map=fields_map,
            )
            new_fields = [x for x in old_fields if x not in remove_fields]
            row["view"]["fields"] = new_fields

        if add_tags:
            row["tags"] = old_tags + [x for x in add_tags if x not in old_tags]

        if remove_tags:
            row["tags"] = [x for x in old_tags if x not in remove_tags]

        if description is not None:
            new_row["description"] = description

        if sort_field is not None:
            sort_field = self.parent.fields.get_field_name(
                value=sort_field, fields_map=fields_map
            )
            new_row["view"]["sort"]["field"] = sort_field

        if sort_descending is not None:
            new_row["view"]["sort"]["desc"] = sort_descending

        if add_column_filters:
            new_column_filters = {}
            new_column_filters.update(old_column_filters)
            new_column_filters.update(add_column_filters)
            new_row["view"]["colFilters"] = new_column_filters

        if remove_column_filters:
            new_column_filters = {
                k: v
                for k, v in old_column_filters.items()
                if k not in remove_column_filters
            }
            new_row["view"]["colFilters"] = new_column_filters

        if gui_page_size:
            gui_page_size = check_gui_page_size(gui_page_size=gui_page_size)

        if row == new_row:
            raise ApiError("No changes supplied!")

        self._update(id=row["uuid"], data=new_row)
        kwargs["value"] = row["uuid"]
        return self.get_by_uuid(**kwargs)

    def update_by_name(self, value, **kwargs):
        """Pass."""
        row = self.get_by_name(value=value, **kwargs)
        kwargs["row"] = row
        self.update(**kwargs)
        return row
    '''

    def delete(self, rows):
        """Delete saved queries returned from get.

        Args:
            rows (:obj:`list` of :obj:`dict`): metadata of saved queries to delete

        Returns:
            :obj:`list` of :obj:`dict`: saved queries deleted
        """
        rows = listify(rows)
        ids = [x["uuid"] for x in rows]
        self._delete(ids=list(set(ids)))
        return rows

    def delete_by_name(self, value, **kwargs):
        """Delete saved queries returned from get.

        Args:
            rows (:obj:`list` of :obj:`dict`): metadata of saved queries to delete

        Returns:
            :obj:`list` of :obj:`dict`: saved queries deleted
        """
        row = self.get_by_name(value=value, **kwargs)
        self.delete(rows=[row])
        return row

    def _add(self, data):
        """Direct API method to create a saved query.

        Warning:
            Queries created with this method will NOT show the filters in the
            query wizard!

        Returns:
            :obj:`str`: ID of the saved query that was created
        """
        path = self.router.views
        return self.request(method="post", path=path, json=data)

    # XXX this is not in public API, need to add
    # def _update(self, id, data):
    #     """Direct API method to update a saved query.

    #     Args:
    #         id (:obj:`str`): id of saved query to update
    #         data (:obj:`dict`): metadata of saved query to update
    #     """
    #     path = self.router.view_by_id.format(id=id)
    #     return self.request(method="post", path=path, json=data)

    def _delete(self, ids):
        """Direct API method to delete saved queries.

        Args:
            ids (:obj:`list` of :obj:`str`): list of saved query uuid's to delete

        Returns:
            :obj:`str`: empty string
        """
        data = {"ids": listify(ids)}
        path = self.router.views
        return self.request(method="delete", path=path, json=data)

    def _get(self, query=None, row_start=0, page_size=None):
        """Direct API method to get saved queries.

        Args:
            query (:obj:`str`, optional): default ``None`` - filter rows to return

                This is NOT a query built by the query wizard!
            row_start (:obj:`int`, optional): default ``0`` - for paging, skip N rows
            page_size (:obj:`int`, optional): default ``0`` - for paging, return N rows

        Returns:
            :obj:`list` of :obj:`dict`: list of saved query metadata
        """
        params = {}
        params["limit"] = page_size
        params["skip"] = row_start
        params["filter"] = query
        path = self.router.views
        return self.request(method="get", path=path, params=params)



# -*- coding: utf-8 -*-
"""Test suite for assets."""
import copy
import io

import pytest

from .callbacks import Callbacks, load_test_data


class CallbacksJson(Callbacks):
    """Pass."""

    @pytest.fixture(scope="class")
    def cbexport(self):
        """Pass."""
        return "json"

    def test_row_as_is(self, cbexport, apiobj):
        """Pass."""
        io_fd = io.StringIO()

        getargs = {"export_fd": io_fd}

        cbobj = self.get_cbobj(apiobj=apiobj, cbexport=cbexport, getargs=getargs)
        cbobj.start()

        start_val = io_fd.getvalue().splitlines()[0]
        assert start_val == "["

        for row in copy.deepcopy(apiobj.TEST_DATA["cb_assets"][:200]):
            row_id = row["internal_axon_id"]
            rows_ret = cbobj.row(row=copy.deepcopy(row))
            assert len(rows_ret) == 1
            assert rows_ret[0] == {"internal_axon_id": row_id}
            assert f'    "internal_axon_id": "{row_id}",' in io_fd.getvalue()

        cbobj.stop()
        stop_val = io_fd.getvalue().splitlines()[-2:]
        assert stop_val == ["]", ""]

    def test_row_fully_loaded(self, cbexport, apiobj):
        """Pass."""
        io_fd = io.StringIO()

        getargs = {
            "export_fd": io_fd,
            "field_excludes": ["adapters"],
            "field_flatten": True,
            "field_titles": True,
            "field_join": True,
            "field_null": True,
            "report_adapters_missing": True,
            "export_schema": True,
        }
        cbobj = self.get_cbobj(apiobj=apiobj, cbexport=cbexport, getargs=getargs)
        cbobj.start()

        start_val = io_fd.getvalue().splitlines()[0]
        assert start_val == "["

        for row in copy.deepcopy(apiobj.TEST_DATA["cb_assets"][:20]):
            row_id = row["internal_axon_id"]
            rows_ret = cbobj.row(row=copy.deepcopy(row))
            assert len(rows_ret) == 1
            assert rows_ret[0] == {"internal_axon_id": row_id}

        cbobj.stop()

        output = io_fd.getvalue()
        assert f'    "Aggregated: Asset Unique ID": "{row_id}",' in output
        assert '"schemas": [' in output
        assert output.endswith("]\n\n")


class TestDevicesCallbacksJson(CallbacksJson):
    """Pass."""

    @pytest.fixture(scope="class")
    def apiobj(self, api_devices):
        """Pass."""
        return load_test_data(apiobj=api_devices)


class TestUsersCallbacksJson(CallbacksJson):
    """Pass."""

    @pytest.fixture(scope="class")
    def apiobj(self, api_users):
        """Pass."""
        return load_test_data(apiobj=api_users)



# -*- coding: utf-8 -*-
"""API models for working with adapters and connections."""
from . import adapters, config, fields, tables

__all__ = (
    "tables",
    "fields",
    "adapters",
    "config",
)



# -*- coding: utf-8 -*-
"""Test suite for axonius_api_client.tools."""
from ....cli import cli
from ....tools import json_load
from ...utils import get_cnx_existing, load_clirunner


class TestGrpCnxCmdGetById:
    """Pass."""

    def test_json_full(self, api_adapters, request, monkeypatch):
        """Pass."""
        cnx = get_cnx_existing(apiobj=api_adapters)

        runner = load_clirunner(request, monkeypatch)

        args1 = [
            "adapters",
            "cnx",
            "get-by-id",
            "--node-name",
            cnx["node_name"],
            "--name",
            cnx["adapter_name"],
            "--id",
            cnx["id"],
            "--export-format",
            "json-full",
        ]
        result1 = runner.invoke(cli=cli, args=args1)

        stderr1 = result1.stderr
        stdout1 = result1.stdout
        exit_code1 = result1.exit_code

        assert stdout1
        assert stderr1
        assert exit_code1 == 0

        json1 = json_load(stdout1)
        assert isinstance(json1, dict)

    def test_json_config(self, api_adapters, request, monkeypatch):
        """Pass."""
        cnx = get_cnx_existing(apiobj=api_adapters)

        runner = load_clirunner(request, monkeypatch)

        args1 = [
            "adapters",
            "cnx",
            "get-by-id",
            "--node-name",
            cnx["node_name"],
            "--name",
            cnx["adapter_name"],
            "--id",
            cnx["id"],
            "--export-format",
            "json-config",
        ]
        result1 = runner.invoke(cli=cli, args=args1)

        stderr1 = result1.stderr
        stdout1 = result1.stdout
        exit_code1 = result1.exit_code

        assert stdout1
        assert stderr1
        assert exit_code1 == 0

        json1 = json_load(stdout1)
        assert isinstance(json1, dict)

    def test_json_basic(self, api_adapters, request, monkeypatch):
        """Pass."""
        cnx = get_cnx_existing(apiobj=api_adapters)

        runner = load_clirunner(request, monkeypatch)

        args1 = [
            "adapters",
            "cnx",
            "get-by-id",
            "--node-name",
            cnx["node_name"],
            "--name",
            cnx["adapter_name"],
            "--id",
            cnx["id"],
            "--export-format",
            "json",
        ]
        result1 = runner.invoke(cli=cli, args=args1)

        stderr1 = result1.stderr
        stdout1 = result1.stdout
        exit_code1 = result1.exit_code

        assert stdout1
        assert stderr1
        assert exit_code1 == 0

        json1 = json_load(stdout1)
        assert isinstance(json1, dict)

    def test_table_schemas(self, api_adapters, request, monkeypatch):
        """Pass."""
        cnx = get_cnx_existing(apiobj=api_adapters)

        runner = load_clirunner(request, monkeypatch)

        args1 = [
            "adapters",
            "cnx",
            "get-by-id",
            "--node-name",
            cnx["node_name"],
            "--name",
            cnx["adapter_name"],
            "--id",
            cnx["id"],
            "--export-format",
            "table-schemas",
        ]

        result1 = runner.invoke(cli=cli, args=args1)

        stderr1 = result1.stderr
        stdout1 = result1.stdout
        exit_code1 = result1.exit_code

        assert stdout1
        assert stderr1
        assert exit_code1 == 0

    def test_table(self, api_adapters, request, monkeypatch):
        """Pass."""
        cnx = get_cnx_existing(apiobj=api_adapters)

        runner = load_clirunner(request, monkeypatch)

        args1 = [
            "adapters",
            "cnx",
            "get-by-id",
            "--node-name",
            cnx["node_name"],
            "--name",
            cnx["adapter_name"],
            "--id",
            cnx["id"],
            "--export-format",
            "table",
        ]

        result1 = runner.invoke(cli=cli, args=args1)

        stderr1 = result1.stderr
        stdout1 = result1.stdout
        exit_code1 = result1.exit_code

        assert stdout1
        assert stderr1
        assert exit_code1 == 0

    def test_str(self, api_adapters, request, monkeypatch):
        """Pass."""
        cnx = get_cnx_existing(apiobj=api_adapters)

        runner = load_clirunner(request, monkeypatch)

        args1 = [
            "adapters",
            "cnx",
            "get-by-id",
            "--node-name",
            cnx["node_name"],
            "--name",
            cnx["adapter_name"],
            "--id",
            cnx["id"],
            "--export-format",
            "str",
        ]

        result1 = runner.invoke(cli=cli, args=args1)

        stderr1 = result1.stderr
        stdout1 = result1.stdout
        exit_code1 = result1.exit_code

        assert stdout1
        assert stderr1
        assert exit_code1 == 0

    def test_str_args(self, api_adapters, request, monkeypatch):
        """Pass."""
        cnx = get_cnx_existing(apiobj=api_adapters)

        runner = load_clirunner(request, monkeypatch)

        args1 = [
            "adapters",
            "cnx",
            "get-by-id",
            "--node-name",
            cnx["node_name"],
            "--name",
            cnx["adapter_name"],
            "--id",
            cnx["id"],
            "--export-format",
            "str-args",
        ]
        result1 = runner.invoke(cli=cli, args=args1)

        stderr1 = result1.stderr
        stdout1 = result1.stdout
        exit_code1 = result1.exit_code

        assert stdout1
        assert stderr1
        assert exit_code1 == 0



# -*- coding: utf-8 -*-
"""Command line interface for Axonius API Client."""
import click

from ...api.entry import Entry
from ...tools import json_dump
from ..options import URL, add_options

EXPORT = click.option(
    "--export-format",
    "-xf",
    "export_format",
    type=click.Choice(["json", "str"]),
    help="Format of to export data in",
    default="str",
    show_envvar=True,
    show_default=True,
)
PASSWORD = click.option(
    "--password",
    "-p",
    "password",
    required=True,
    help="Password to assign to admin user",
    prompt="Password to assign to admin user",
    hide_input=True,
    show_envvar=True,
    show_default=True,
)

COMPANY = click.option(
    "--company-name",
    "-cn",
    "company_name",
    required=True,
    help="Company Name",
    prompt="Company Name",
    show_envvar=True,
    show_default=True,
)

CONTACT = click.option(
    "--contact-email",
    "-ce",
    "contact_email",
    required=True,
    help="Contact Email",
    prompt="Contact Email",
    show_envvar=True,
    show_default=True,
)

OPTIONS = [URL, PASSWORD, COMPANY, CONTACT, EXPORT]


@click.command(name="signup")
@add_options(OPTIONS)
@click.pass_context
def cmd(ctx, url, password, company_name, contact_email, export_format):
    """Perform the initial signup to an instance."""
    entry = Entry(url=url)
    with ctx.obj.exc_wrap(wraperror=ctx.obj.wraperror):
        data = entry.signup(
            password=password, company_name=company_name, contact_email=contact_email
        )

    if export_format == "str":
        lines = [
            f"AX_URL={url}",
            f"AX_KEY={data['api_key']}",
            f"AX_SECRET={data['api_secret']}",
        ]
        click.secho("\n".join(lines))
        ctx.exit(0)

    if export_format == "json":
        data["url"] = url
        click.secho(json_dump(data))

    ctx.exit(1)



# -*- coding: utf-8 -*-
"""Command line interface for Axonius API Client."""
from ...context import CONTEXT_SETTINGS, click
from ...options import AUTH, add_options
from .grp_common import ROLE_NAME

OPTIONS = [*AUTH, ROLE_NAME]


@click.command(name="delete", context_settings=CONTEXT_SETTINGS)
@add_options(OPTIONS)
@click.pass_context
def cmd(ctx, url, key, secret, name, **kwargs):
    """Delete a role."""
    client = ctx.obj.start_client(url=url, key=key, secret=secret)

    with ctx.obj.exc_wrap(wraperror=ctx.obj.wraperror):
        client.system.roles.delete(name=name)
        ctx.obj.echo_ok(f"Deleted role {name!r}")



# -*- coding: utf-8 -*-
"""Command line interface for Axonius API Client."""
from ....constants import GUI_PAGE_SIZES
from ....tools import json_dump, listify
from ...context import SplitEquals, click
from ...options import int_callback

EXPORT_FORMAT = click.option(
    "--export-format",
    "-xf",
    "export_format",
    type=click.Choice(["json", "str", "str-names"]),
    help="Format of to export data in",
    default="str",
    show_envvar=True,
    show_default=True,
)

SQ_OPTS = [
    click.option(
        "--tag",
        "-t",
        "tags",
        help="Tags to set for saved query",
        multiple=True,
        show_envvar=True,
        show_default=True,
    ),
    click.option(
        "--sort-field",
        "-sf",
        "sort_field",
        help="Column to sort data on.",
        metavar="ADAPTER:FIELD",
        show_envvar=True,
        show_default=True,
    ),
    click.option(
        "--sort-ascending",
        "-sd",
        "sort_descending",
        default=True,
        help="Sort --sort-field ascending.",
        is_flag=True,
        show_envvar=True,
    ),
    click.option(
        "--column-filter",
        "-cf",
        "column_filters",
        help="Columns to filter in the format of adapter:field=value.",
        metavar="ADAPTER:FIELD=value",
        type=SplitEquals(),
        multiple=True,
        show_envvar=True,
    ),
    click.option(
        "--gui-page-size",
        "-gps",
        default=format(GUI_PAGE_SIZES[0]),
        help="Number of rows to show per page in GUI.",
        type=click.Choice([format(x) for x in GUI_PAGE_SIZES]),
        callback=int_callback,
        show_envvar=True,
        show_default=True,
    ),
    click.option(
        "--description",
        "-d",
        "description",
        help="Description to set on saved query",
        show_envvar=True,
        show_default=True,
        default=None,
    ),
]


def handle_export(ctx, rows, export_format):
    """Pass."""
    if export_format == "json":
        click.secho(json_dump(rows))
        ctx.exit(0)

    if export_format == "str-names":
        names = "\n".join([x["name"] for x in listify(rows)])
        click.secho(names)
        ctx.exit(0)

    if export_format == "str":
        for row in listify(rows):
            name = row["name"]
            description = row.get("description", "")
            query = row["view"].get("query", {}).get("filter", None)

            tags = "\n  " + "\n  ".join(row.get("tags", []))
            fields = "\n  " + "\n  ".join(row["view"].get("fields", []))

            click.secho("\n-----------------------------------------------")
            click.secho(f"Name: {name}")
            click.secho(f"Description: {description}")
            click.secho(f"Query: {query}")
            click.secho(f"Tags: {tags}")
            click.secho(f"Fields: {fields}")

        ctx.exit(0)



# -*- coding: utf-8 -*-
"""Command line interface for Axonius API Client."""
from ....tools import json_dump
from ...context import click

EXPORT = click.option(
    "--export-format",
    "-xf",
    "export_format",
    type=click.Choice(["json-full", "json", "str"]),
    help="Format of to export data in",
    default="str",
    show_envvar=True,
    show_default=True,
)


def echo_key(data, key):
    """Pass."""
    value = data[key]

    key = key.replace("_", " ").title()

    if isinstance(value, list):
        value = ", ".join(value)

    click.secho(f"{key}: {value}")


def echo_fetch_progress(data, key):
    """Pass."""
    phases = data["phases"]
    fetch = [x for x in phases if x["name"] == key][0]
    key = key.replace("_", " ").title()

    progress = fetch["progress"]
    for status, adapters in progress.items():
        adapters = ", ".join(adapters)
        click.secho(f"Phase {key} {status}: {adapters}\n")


def handle_export(ctx, data, export_format, **kwargs):
    """Pass."""
    if export_format == "json":
        data.pop("phases")
        click.secho(json_dump(data))
        ctx.exit(0)

    if export_format == "json-full":
        click.secho(json_dump(data))
        ctx.exit(0)

    if export_format == "str":
        echo_fetch_progress(data, "Fetch_Devices")
        echo_fetch_progress(data, "Fetch_Scanners")
        keys = [
            "last_start_date",
            "last_finish_date",
            "last_took_minutes",
            "next_start_date",
            "next_in_minutes",
            "is_running",
            "phases_done",
            "phases_pending",
        ]
        for key in keys:
            echo_key(data, key)

        # adapters fetching
        # adapters done
        # echo_key(data, last_start)
        # data.pop("phases")
        # for k, v in data.items():
        #     k = k.replace("_", " ").title()
        #     if isinstance(v, list):
        #         v = ", ".join(v)
        #     click.secho(f"{k}: {v}")
        ctx.exit(0)

    ctx.exit(1)



# -*- coding: utf-8 -*-
"""API models for working with device and user assets."""
import ipaddress

from ..routers import API_VERSION
from .asset_mixin import AssetMixin


class Devices(AssetMixin):
    """Device related API methods."""

    FIELD_HOSTNAME = "specific_data.data.hostname"
    FIELD_IP = "specific_data.data.network_interfaces.ips"
    FIELD_IP_RAW = "specific_data.data.network_interfaces.ips_raw"
    FIELD_MAC = "specific_data.data.network_interfaces.mac"
    FIELD_SUBNET = "specific_data.data.network_interfaces.subnets"

    @property
    def fields_default(self):
        """Fields to use by default for getting assets."""
        return self.FIELDS_API + [
            self.FIELD_HOSTNAME,
            self.FIELD_IP,
            self.FIELD_MAC,
            self.FIELD_SUBNET,
        ]

    @property
    def router(self):
        """Router for this API model."""
        return API_VERSION.devices

    def get_by_hostnames(self, values, **kwargs):
        """Build a query to get assets where hostname in values."""
        kwargs["field"] = self.FIELD_HOSTNAME
        kwargs["field_manual"] = True
        kwargs["values"] = values
        return self.get_by_values(**kwargs)

    def get_by_hostname_regex(self, value, **kwargs):
        """Build a query to get assets where hostname regex matches value."""
        kwargs["field"] = self.FIELD_HOSTNAME
        kwargs["field_manual"] = True
        kwargs["value"] = value
        return self.get_by_value_regex(**kwargs)

    def get_by_hostname(self, value, **kwargs):
        """Build a query to get assets where hostname == value."""
        kwargs["field"] = self.FIELD_HOSTNAME
        kwargs["field_manual"] = True
        kwargs["value"] = value
        return self.get_by_value(**kwargs)

    def get_by_macs(self, values, **kwargs):
        """Build a query to get assets where mac in values."""
        kwargs["field"] = self.FIELD_MAC
        kwargs["field_manual"] = True
        kwargs["values"] = values
        return self.get_by_values(**kwargs)

    def get_by_mac_regex(self, value, **kwargs):
        """Build a query to get assets where mac regex matches value."""
        kwargs["field"] = self.FIELD_MAC
        kwargs["field_manual"] = True
        kwargs["value"] = value
        return self.get_by_value_regex(**kwargs)

    def get_by_mac(self, value, **kwargs):
        """Build a query to get assets where mac == value."""
        kwargs["field"] = self.FIELD_MAC
        kwargs["field_manual"] = True
        kwargs["value"] = value
        return self.get_by_value(**kwargs)

    def get_by_ips(self, values, **kwargs):
        """Build a query to get assets where ip in values."""
        kwargs["field"] = self.FIELD_IP
        kwargs["field_manual"] = True
        kwargs["values"] = values
        return self.get_by_values(**kwargs)

    def get_by_ip_regex(self, value, **kwargs):
        """Build a query to get assets where ip regex matches value."""
        kwargs["field"] = self.FIELD_IP
        kwargs["field_manual"] = True
        kwargs["value"] = value
        return self.get_by_value_regex(**kwargs)

    def get_by_ip(self, value, **kwargs):
        """Build a query to get assets where ip == value."""
        kwargs["field"] = self.FIELD_IP
        kwargs["field_manual"] = True
        kwargs["value"] = value
        return self.get_by_value(**kwargs)

    def get_by_subnet(self, value, not_flag=False, pre="", post="", **kwargs):
        """Build a query to get assets where ip in subnet.

        Args:
            value (:obj:`str`): value to that must match field
                "network_interfaces.ips_raw"
            **kwargs: passed to :meth:`AssetMixin.get`

        Yields:
            :obj:`dict`: asset matching query that is built if generator is True

        Returns:
            :obj:`list` of :obj:`dict`: assets matching query that is built
                if generator is False
        """
        field = self.FIELD_IP_RAW

        network = ipaddress.ip_network(value)
        gte = int(network.network_address)
        lte = int(network.broadcast_address)
        match = "".join(['match({"$gte": ', str(gte), ', "$lte": ', str(lte), "})"])

        inner = f"{field} == {match}"

        kwargs["query"] = self._build_query(
            inner=inner, pre=pre, post=post, not_flag=not_flag,
        )

        return self.get(**kwargs)



# -*- coding: utf-8 -*-
"""Command line interface for Axonius API Client."""
import pathlib

import click
import OpenSSL

from ..options import add_options

PEM_TYPE = OpenSSL.crypto.FILETYPE_PEM
ASN1_TYPE = OpenSSL.crypto.FILETYPE_ASN1

PATH = click.option(
    "--path",
    "-p",
    "path",
    type=click.Path(exists=True, dir_okay=False, readable=True, resolve_path=False),
    help="Path to SSL certificate to convert from binary to Base64",
    show_envvar=True,
    show_default=True,
    required=True,
)

OPTIONS = [PATH]


@click.command(name="convert-cert")
@add_options(OPTIONS)
@click.pass_context
def cmd(ctx, path):
    """Convert an SSL Certificate from binary to Base64."""
    path = pathlib.Path(path).expanduser().resolve()
    contents = path.read_bytes()
    der = load_der(contents=contents)
    pem = der_to_pem(der=der)
    write_pem_path(ctx=ctx, path=path, pem=pem)
    ctx.exit(0)


def load_der(contents):
    """Load the bytes DER cert file into a python object."""
    return OpenSSL.crypto.load_certificate(ASN1_TYPE, contents)


def der_to_pem(der):
    """Convert a binary bytes DER cert to ascii PEM cert."""
    return OpenSSL.crypto.dump_certificate(PEM_TYPE, der)


def write_pem_path(ctx, path, pem):
    """Write an ascii PEM file to a path."""
    base_name = path.stem
    parent = path.parent
    pem_name = f"{base_name}.pem"
    pem_path = parent / pem_name
    if pem_path.is_file():
        ctx.obj.echo_error(f"Base64 SSL Certificate already exists: {pem_path}")
    pem_path.write_bytes(pem)
    ctx.obj.echo_ok(f"Wrote Base64 SSL Certificate to: {pem_path}")
    return pem_path



# -*- coding: utf-8 -*-
"""Command line interface for Axonius API Client."""
from ...tools import json_dump
from ..context import CONTEXT_SETTINGS, click
from ..options import AUTH, add_options

OPTIONS = [
    *AUTH,
    click.option(
        "--id",
        "-i",
        "id",
        help="Internal Axon ID of asset",
        required=True,
        show_envvar=True,
        show_default=True,
    ),
]


@click.command(name="get-by-id", context_settings=CONTEXT_SETTINGS)
@add_options(OPTIONS)
@click.pass_context
def cmd(ctx, url, key, secret, **kwargs):
    """Get ALL data for a single asset."""
    client = ctx.obj.start_client(url=url, key=key, secret=secret)

    p_grp = ctx.parent.command.name
    apiobj = getattr(client, p_grp)

    with ctx.obj.exc_wrap(wraperror=ctx.obj.wraperror):
        rows = apiobj.get_by_id(**kwargs)

    click.secho(json_dump(rows))
    ctx.exit(0)



# -*- coding: utf-8 -*-
"""Command line interface for Axonius API Client."""
import click

from ...context import AliasedGroup
from . import cmd_get, cmd_restore_from_aws_s3, cmd_update


@click.group(cls=AliasedGroup)
def central_core():
    """Group: Manage Central Core feature."""


central_core.add_command(cmd_get.cmd)
central_core.add_command(cmd_update.cmd)
central_core.add_command(cmd_restore_from_aws_s3.cmd)
